# Copyright (c) 2025 Airbyte, Inc., all rights reserved.
"""Utility and factory functions for testing Airbyte connectors."""

from pathlib import Path
from typing import Any, Literal

import yaml

from airbyte_cdk.test.standard_tests.connector_base import ConnectorTestSuiteBase
from airbyte_cdk.test.standard_tests.declarative_sources import (
    DeclarativeSourceTestSuite,
)
from airbyte_cdk.test.standard_tests.destination_base import DestinationTestSuiteBase
from airbyte_cdk.test.standard_tests.docker_base import DockerConnectorTestSuite
from airbyte_cdk.test.standard_tests.source_base import SourceTestSuiteBase
from airbyte_cdk.utils.connector_paths import (
    METADATA_YAML,
    find_connector_root_from_name,
)


def create_connector_test_suite(
    *,
    connector_name: str | None = None,
    connector_directory: Path | None = None,
) -> type[ConnectorTestSuiteBase]:
    """Get the test class for the specified connector name or path."""
    if connector_name and connector_directory:
        raise ValueError("Specify either `connector_name` or `connector_directory`, not both.")
    if not connector_name and not connector_directory:
        raise ValueError("Specify either `connector_name` or `connector_directory`.")

    if connector_name:
        connector_directory = find_connector_root_from_name(
            connector_name,
        )
    else:
        # By here, we know that connector_directory is not None
        # but connector_name is None. Set the connector_name.
        assert connector_directory is not None, "connector_directory should not be None here."
        connector_name = connector_directory.absolute().name

    metadata_yaml_path = connector_directory / METADATA_YAML
    if not metadata_yaml_path.exists():
        raise FileNotFoundError(
            f"Could not find metadata YAML file '{metadata_yaml_path}' relative to the connector directory."
        )
    metadata_dict: dict[str, Any] = yaml.safe_load(metadata_yaml_path.read_text())
    metadata_tags = metadata_dict["data"].get("tags", [])
    language_tags: list[str] = [tag for tag in metadata_tags if tag.startswith("language:")]
    if not language_tags:
        raise ValueError(
            f"Metadata YAML file '{metadata_yaml_path}' does not contain a 'language' tag. "
            "Please ensure the metadata file is correctly configured."
            f"Found tags: {', '.join(metadata_tags)}"
        )
    language = language_tags[0].split(":")[1]

    if language == "java":
        test_suite_class = DockerConnectorTestSuite
    elif language == "manifest-only":
        test_suite_class = DeclarativeSourceTestSuite
    elif language == "python" and connector_name.startswith("source-"):
        test_suite_class = SourceTestSuiteBase
    elif language == "python" and connector_name.startswith("destination-"):
        test_suite_class = DestinationTestSuiteBase
    else:
        raise ValueError(f"Unsupported language for connector '{connector_name}': {language}")

    subclass_overrides: dict[str, Any] = {
        "get_connector_root_dir": classmethod(lambda cls: connector_directory),
    }

    TestSuiteAuto = type(
        "TestSuiteAuto",
        (test_suite_class,),
        subclass_overrides,
    )

    return TestSuiteAuto



#
# Copyright (c) 2025 Airbyte, Inc., all rights reserved.
#


from airbyte_cdk.manifest_migrations.manifest_migration import (
    TYPE_TAG,
    ManifestMigration,
    ManifestType,
)


class HttpRequesterRequestBodyJsonDataToRequestBody(ManifestMigration):
    """
    This migration is responsible for migrating the `request_body_json` and `request_body_data` keys
    to a unified `request_body` key in the HttpRequester component.
    The migration will copy the value of either original key to `request_body` and remove the original key.
    """

    component_type = "HttpRequester"

    body_json_key = "request_body_json"
    body_data_key = "request_body_data"
    original_keys = (body_json_key, body_data_key)

    replacement_key = "request_body"

    def should_migrate(self, manifest: ManifestType) -> bool:
        return manifest[TYPE_TAG] == self.component_type and any(
            key in list(manifest.keys()) for key in self.original_keys
        )

    def migrate(self, manifest: ManifestType) -> None:
        for key in self.original_keys:
            if key == self.body_json_key and key in manifest:
                self._migrate_body_json(manifest, key)
            elif key == self.body_data_key and key in manifest:
                self._migrate_body_data(manifest, key)

    def validate(self, manifest: ManifestType) -> bool:
        return self.replacement_key in manifest and all(
            key not in manifest for key in self.original_keys
        )

    def _migrate_body_json(self, manifest: ManifestType, key: str) -> None:
        """
        Migrate the value of the request_body_json.
        """
        query_key = "query"
        text_type = "RequestBodyPlainText"
        graph_ql_type = "RequestBodyGraphQL"
        json_object_type = "RequestBodyJsonObject"

        if isinstance(manifest[key], str):
            self._migrate_value(manifest, key, text_type)
        elif isinstance(manifest[key], dict):
            if isinstance(manifest[key].get(query_key), str):
                self._migrate_value(manifest, key, graph_ql_type)
            else:
                self._migrate_value(manifest, key, json_object_type)

    def _migrate_body_data(self, manifest: ManifestType, key: str) -> None:
        """
        Migrate the value of the request_body_data.
        """
        self._migrate_value(manifest, key, "RequestBodyUrlEncodedForm")

    def _migrate_value(self, manifest: ManifestType, key: str, type: str) -> None:
        """
        Migrate the value of the key to a specific type and update the manifest.
        """
        manifest[self.replacement_key] = {
            "type": type,
            "value": manifest[key],
        }
        manifest.pop(key, None)



import os
from hashlib import md5
from pathlib import Path
from typing import Any, cast

import yaml
from boltons.typeutils import classproperty

from airbyte_cdk.sources.declarative.concurrent_declarative_source import (
    ConcurrentDeclarativeSource,
)
from airbyte_cdk.test.models import ConnectorTestScenario
from airbyte_cdk.test.standard_tests._job_runner import IConnector
from airbyte_cdk.test.standard_tests.source_base import SourceTestSuiteBase
from airbyte_cdk.utils.connector_paths import MANIFEST_YAML


def md5_checksum(file_path: Path) -> str:
    """Helper function to calculate the MD5 checksum of a file.

    This is used to calculate the checksum of the `components.py` file, if it exists.
    """
    with open(file_path, "rb") as file:
        return md5(file.read()).hexdigest()


class DeclarativeSourceTestSuite(SourceTestSuiteBase):
    """Declarative source test suite.

    This inherits from the Python-based source test suite and implements the
    `create_connector` method to create a declarative source object instead of
    requiring a custom Python source object.

    The class also automatically locates the `manifest.yaml` file and the
    `components.py` file (if it exists) in the connector's directory.
    """

    connector: type[IConnector] | None = None

    @classproperty
    def manifest_yaml_path(cls) -> Path:
        """Get the path to the manifest.yaml file."""
        result = cls.get_connector_root_dir() / MANIFEST_YAML
        if result.exists():
            return result

        raise FileNotFoundError(
            f"Manifest YAML file not found at {result}. "
            "Please ensure that the test suite is run in the correct directory.",
        )

    @classproperty
    def components_py_path(cls) -> Path | None:
        """Get the path to the `components.py` file, if one exists.

        If not `components.py` file exists, return None.
        """
        result = cls.get_connector_root_dir() / "components.py"
        if result.exists():
            return result

        return None

    @classmethod
    def create_connector(
        cls,
        scenario: ConnectorTestScenario | None,
    ) -> IConnector:
        """Create a connector scenario for the test suite.

        This overrides `create_connector` from the create a declarative source object
        instead of requiring a custom python source object.

        Subclasses should not need to override this method.
        """
        scenario = scenario or ConnectorTestScenario()  # Use default (empty) scenario if None
        manifest_dict = yaml.safe_load(cls.manifest_yaml_path.read_text())
        config = {
            "__injected_manifest": manifest_dict,
        }
        config.update(
            scenario.get_config_dict(
                empty_if_missing=True,
                connector_root=cls.get_connector_root_dir(),
            ),
        )

        if cls.components_py_path and cls.components_py_path.exists():
            os.environ["AIRBYTE_ENABLE_UNSAFE_CODE"] = "true"
            config["__injected_components_py"] = cls.components_py_path.read_text()
            config["__injected_components_py_checksums"] = {
                "md5": md5_checksum(cls.components_py_path),
            }

        return cast(
            IConnector,
            ConcurrentDeclarativeSource(
                config=config,
                catalog=None,
                state=None,
                source_config=manifest_dict,
            ),
        )



#
# Copyright (c) 2025 Airbyte, Inc., all rights reserved.
#

from dataclasses import dataclass
from typing import Any

from airbyte_cdk.sources.declarative.validators.validation_strategy import ValidationStrategy


@dataclass
class PredicateValidator:
    """
    Validator that applies a validation strategy to a value.
    """

    value: Any
    strategy: ValidationStrategy

    def validate(self) -> None:
        """
        Applies the validation strategy to the value.

        :raises ValueError: If validation fails
        """
        self.strategy.validate(self.value)



from typing import Any, Mapping, Optional

from airbyte_cdk.sources.declarative.schema import SchemaLoader


class CachingSchemaLoaderDecorator(SchemaLoader):
    def __init__(self, schema_loader: SchemaLoader):
        self._decorated = schema_loader
        self._loaded_schema: Optional[Mapping[str, Any]] = None

    def get_json_schema(self) -> Mapping[str, Any]:
        if self._loaded_schema is None:
            self._loaded_schema = self._decorated.get_json_schema()

        return self._loaded_schema  # type: ignore  # at that point, we assume the schema will be populated



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#


import copy
import json
import logging
from dataclasses import InitVar, dataclass
from typing import (
    TYPE_CHECKING,
    Any,
    Iterable,
    List,
    Mapping,
    MutableMapping,
    Optional,
    TypeVar,
    Union,
)

import dpath
import requests

from airbyte_cdk.models import AirbyteMessage
from airbyte_cdk.sources.declarative.interpolation.interpolated_string import InterpolatedString
from airbyte_cdk.sources.declarative.partition_routers.partition_router import PartitionRouter
from airbyte_cdk.sources.declarative.requesters.request_option import (
    RequestOption,
    RequestOptionType,
)
from airbyte_cdk.sources.streams.concurrent.partitions.partition import Partition
from airbyte_cdk.sources.types import Config, StreamSlice, StreamState

if TYPE_CHECKING:
    from airbyte_cdk.sources.streams.concurrent.abstract_stream import AbstractStream


T = TypeVar("T")


def iterate_with_last_flag(generator: Iterable[T]) -> Iterable[tuple[T | None, bool]]:
    iterator = iter(generator)

    try:
        current = next(iterator)
    except StopIteration:
        yield None, True
        return  # Return an empty iterator

    for next_item in iterator:
        yield current, False
        current = next_item

    yield current, True


@dataclass
class ParentStreamConfig:
    """
    Describes how to create a stream slice from a parent stream

    stream: The stream to read records from
    parent_key: The key of the parent stream's records that will be the stream slice key
    partition_field: The partition key
    extra_fields: Additional field paths to include in the stream slice
    request_option: How to inject the slice value on an outgoing HTTP request
    incremental_dependency (bool): Indicates if the parent stream should be read incrementally.
    """

    stream: "AbstractStream"
    parent_key: Union[InterpolatedString, str]
    partition_field: Union[InterpolatedString, str]
    config: Config
    parameters: InitVar[Mapping[str, Any]]
    extra_fields: Optional[Union[List[List[str]], List[List[InterpolatedString]]]] = (
        None  # List of field paths (arrays of strings)
    )
    request_option: Optional[RequestOption] = None
    incremental_dependency: bool = False
    lazy_read_pointer: Optional[List[Union[InterpolatedString, str]]] = None

    def __post_init__(self, parameters: Mapping[str, Any]) -> None:
        self.parent_key = InterpolatedString.create(self.parent_key, parameters=parameters)
        self.partition_field = InterpolatedString.create(
            self.partition_field, parameters=parameters
        )
        if self.extra_fields:
            # Create InterpolatedString for each field path in extra_keys
            self.extra_fields = [
                [InterpolatedString.create(path, parameters=parameters) for path in key_path]
                for key_path in self.extra_fields
            ]

        self.lazy_read_pointer = (
            [
                InterpolatedString.create(path, parameters=parameters)
                if isinstance(path, str)
                else path
                for path in self.lazy_read_pointer
            ]
            if self.lazy_read_pointer
            else None
        )


@dataclass
class SubstreamPartitionRouter(PartitionRouter):
    """
    Partition router that iterates over the parent's stream records and emits slices
    Will populate the state with `partition_field` and `parent_slice` so they can be accessed by other components

    Attributes:
        parent_stream_configs (List[ParentStreamConfig]): parent streams to iterate over and their config
    """

    parent_stream_configs: List[ParentStreamConfig]
    config: Config
    parameters: InitVar[Mapping[str, Any]]

    def __post_init__(self, parameters: Mapping[str, Any]) -> None:
        if not self.parent_stream_configs:
            raise ValueError("SubstreamPartitionRouter needs at least 1 parent stream")
        self._parameters = parameters

    def get_request_params(
        self,
        stream_state: Optional[StreamState] = None,
        stream_slice: Optional[StreamSlice] = None,
        next_page_token: Optional[Mapping[str, Any]] = None,
    ) -> Mapping[str, Any]:
        # Pass the stream_slice from the argument, not the cursor because the cursor is updated after processing the response
        return self._get_request_option(RequestOptionType.request_parameter, stream_slice)

    def get_request_headers(
        self,
        stream_state: Optional[StreamState] = None,
        stream_slice: Optional[StreamSlice] = None,
        next_page_token: Optional[Mapping[str, Any]] = None,
    ) -> Mapping[str, Any]:
        # Pass the stream_slice from the argument, not the cursor because the cursor is updated after processing the response
        return self._get_request_option(RequestOptionType.header, stream_slice)

    def get_request_body_data(
        self,
        stream_state: Optional[StreamState] = None,
        stream_slice: Optional[StreamSlice] = None,
        next_page_token: Optional[Mapping[str, Any]] = None,
    ) -> Mapping[str, Any]:
        # Pass the stream_slice from the argument, not the cursor because the cursor is updated after processing the response
        return self._get_request_option(RequestOptionType.body_data, stream_slice)

    def get_request_body_json(
        self,
        stream_state: Optional[StreamState] = None,
        stream_slice: Optional[StreamSlice] = None,
        next_page_token: Optional[Mapping[str, Any]] = None,
    ) -> Mapping[str, Any]:
        # Pass the stream_slice from the argument, not the cursor because the cursor is updated after processing the response
        return self._get_request_option(RequestOptionType.body_json, stream_slice)

    def _get_request_option(
        self, option_type: RequestOptionType, stream_slice: Optional[StreamSlice]
    ) -> Mapping[str, Any]:
        params: MutableMapping[str, Any] = {}
        if stream_slice:
            for parent_config in self.parent_stream_configs:
                if (
                    parent_config.request_option
                    and parent_config.request_option.inject_into == option_type
                ):
                    key = parent_config.partition_field.eval(self.config)  # type: ignore # partition_field is always casted to an interpolated string
                    value = stream_slice.get(key)
                    if value:
                        parent_config.request_option.inject_into_request(params, value, self.config)
        return params

    def stream_slices(self) -> Iterable[StreamSlice]:
        """
        Iterate over each parent stream's record and create a StreamSlice for each record.

        For each stream, iterate over its stream_slices.
        For each stream slice, iterate over each record.
        yield a stream slice for each such records.

        If a parent slice contains no record, emit a slice with parent_record=None.

        The template string can interpolate the following values:
        - parent_stream_slice: mapping representing the parent's stream slice
        - parent_record: mapping representing the parent record
        - parent_stream_name: string representing the parent stream name
        """
        if not self.parent_stream_configs:
            yield from []
        else:
            for parent_stream_config in self.parent_stream_configs:
                parent_stream = parent_stream_config.stream
                parent_field = parent_stream_config.parent_key.eval(self.config)  # type: ignore # parent_key is always casted to an interpolated string
                partition_field = parent_stream_config.partition_field.eval(self.config)  # type: ignore # partition_field is always casted to an interpolated string
                extra_fields = None
                if parent_stream_config.extra_fields:
                    extra_fields = [
                        [field_path_part.eval(self.config) for field_path_part in field_path]  # type: ignore [union-attr]
                        for field_path in parent_stream_config.extra_fields
                    ]

                for partition, is_last_slice in iterate_with_last_flag(
                    parent_stream.generate_partitions()
                ):
                    if partition is None:
                        break
                    for parent_record, is_last_record_in_slice in iterate_with_last_flag(
                        partition.read()
                    ):
                        emit_slice = parent_record is not None
                        if parent_record is not None:
                            # In the previous CDK implementation, state management was done internally by the stream.
                            # However, this could cause issues when doing availability check for example as the availability
                            # check would progress the state so state management was moved outside of the read method.
                            # Hence, we need to call the cursor here.
                            # Note that we call observe and close_partition before emitting the associated record as the
                            # ConcurrentPerPartitionCursor will associate a record with the state of the stream after the
                            # record was consumed.
                            parent_stream.cursor.observe(parent_record)
                            parent_partition = (
                                parent_record.associated_slice.partition
                                if parent_record.associated_slice
                                else {}
                            )
                            record_data = parent_record.data

                            try:
                                partition_value = dpath.get(
                                    record_data,  # type: ignore [arg-type]
                                    parent_field,
                                )
                            except KeyError:
                                # FIXME a log here would go a long way for debugging
                                emit_slice = False

                            if emit_slice:
                                # Add extra fields
                                extracted_extra_fields = self._extract_extra_fields(
                                    record_data, extra_fields
                                )

                                if parent_stream_config.lazy_read_pointer:
                                    extracted_extra_fields = {
                                        "child_response": self._extract_child_response(
                                            record_data,
                                            parent_stream_config.lazy_read_pointer,  # type: ignore[arg-type]  # lazy_read_pointer type handeled in __post_init__ of parent_stream_config
                                        ),
                                        **extracted_extra_fields,
                                    }

                        if is_last_record_in_slice:
                            parent_stream.cursor.close_partition(partition)
                            if is_last_slice:
                                parent_stream.cursor.ensure_at_least_one_state_emitted()

                        if emit_slice:
                            yield StreamSlice(
                                partition={
                                    partition_field: partition_value,
                                    "parent_slice": parent_partition or {},
                                },
                                cursor_slice={},
                                extra_fields=extracted_extra_fields,
                            )

                yield from []

    def _extract_child_response(
        self, parent_record: Mapping[str, Any] | AirbyteMessage, pointer: List[InterpolatedString]
    ) -> requests.Response:
        """Extract child records from a parent record based on lazy pointers."""

        def _create_response(data: MutableMapping[str, Any]) -> SafeResponse:
            """Create a SafeResponse with the given data."""
            response = SafeResponse()
            response.content = json.dumps(data).encode("utf-8")
            response.status_code = 200
            return response

        path = [path.eval(self.config) for path in pointer]
        return _create_response(dpath.get(parent_record, path, default=[]))  # type: ignore # argunet will be a MutableMapping, given input data structure

    def _extract_extra_fields(
        self,
        parent_record: Mapping[str, Any] | AirbyteMessage,
        extra_fields: Optional[List[List[str]]] = None,
    ) -> Mapping[str, Any]:
        """
        Extracts additional fields specified by their paths from the parent record.

        Args:
            parent_record (Mapping[str, Any]): The record from the parent stream to extract fields from.
            extra_fields (Optional[List[List[str]]]): A list of field paths (as lists of strings) to extract from the parent record.

        Returns:
            Mapping[str, Any]: A dictionary containing the extracted fields.
                               The keys are the joined field paths, and the values are the corresponding extracted values.
        """
        extracted_extra_fields = {}
        if extra_fields:
            for extra_field_path in extra_fields:
                try:
                    extra_field_value = dpath.get(
                        parent_record,  # type: ignore [arg-type]
                        extra_field_path,
                    )
                    self.logger.debug(
                        f"Extracted extra_field_path: {extra_field_path} with value: {extra_field_value}"
                    )
                except KeyError:
                    self.logger.debug(f"Failed to extract extra_field_path: {extra_field_path}")
                    extra_field_value = None
                extracted_extra_fields[".".join(extra_field_path)] = extra_field_value
        return extracted_extra_fields

    def _migrate_child_state_to_parent_state(self, stream_state: StreamState) -> StreamState:
        """
        Migrate the child or global stream state into the parent stream's state format.

        This method converts the child stream state—or, if present, the global state—into a format that is
        compatible with parent streams that use incremental synchronization. The migration occurs only for
        parent streams with incremental dependencies. It filters out per-partition states and retains only the
        global state in the form {cursor_field: cursor_value}.

        The method supports multiple input formats:
          - A simple global state, e.g.:
                {"updated_at": "2023-05-27T00:00:00Z"}
          - A state object that contains a "state" key (which is assumed to hold the global state), e.g.:
                {"state": {"updated_at": "2023-05-27T00:00:00Z"}, ...}
            In this case, the migration uses the first value from the "state" dictionary.
          - Any per-partition state formats or other non-simple structures are ignored during migration.

        Args:
            stream_state (StreamState): The state to migrate. Expected formats include:
                - {"updated_at": "2023-05-27T00:00:00Z"}
                - {"state": {"updated_at": "2023-05-27T00:00:00Z"}, ...}
                  (In this format, only the first global state value is used, and per-partition states are ignored.)

        Returns:
            StreamState: A migrated state for parent streams in the format:
                {
                    "parent_stream_name": {"parent_stream_cursor": "2023-05-27T00:00:00Z"}
                }
            where each parent stream with an incremental dependency is assigned its corresponding cursor value.

        Example:
            Input: {"updated_at": "2023-05-27T00:00:00Z"}
            Output: {
                "parent_stream_name": {"parent_stream_cursor": "2023-05-27T00:00:00Z"}
            }
        """
        substream_state_values = list(stream_state.values())
        substream_state = substream_state_values[0] if substream_state_values else {}

        # Ignore per-partition states or invalid formats.
        if isinstance(substream_state, (list, dict)) or len(substream_state_values) != 1:
            # If a global state is present under the key "state", use its first value.
            if (
                "state" in stream_state
                and isinstance(stream_state["state"], dict)
                and stream_state["state"] != {}
            ):
                substream_state = list(stream_state["state"].values())[0]
            else:
                return {}

        # Build the parent state for all parent streams with incremental dependencies.
        parent_state = {}
        if substream_state:
            for parent_config in self.parent_stream_configs:
                if parent_config.incremental_dependency:
                    parent_state[parent_config.stream.name] = {
                        parent_config.stream.cursor_field: substream_state
                    }

        return parent_state

    def get_stream_state(self) -> Optional[Mapping[str, StreamState]]:
        """
        Get the state of the parent streams.

        Returns:
            StreamState: The current state of the parent streams.

        Example of state format:
        {
            "parent_stream_name1": {
                "last_updated": "2023-05-27T00:00:00Z"
            },
            "parent_stream_name2": {
                "last_updated": "2023-05-27T00:00:00Z"
            }
        }
        """
        parent_state = {}
        for parent_config in self.parent_stream_configs:
            if parent_config.incremental_dependency:
                parent_state[parent_config.stream.name] = copy.deepcopy(
                    parent_config.stream.cursor.state
                )
        return parent_state

    @property
    def logger(self) -> logging.Logger:
        return logging.getLogger("airbyte.SubstreamPartitionRouter")


class SafeResponse(requests.Response):
    """
    A subclass of requests.Response that acts as an interface to migrate parsed child records
    into a response object. This allows seamless interaction with child records as if they
    were original response, ensuring compatibility with methods that expect requests.Response data type.
    """

    def __getattr__(self, name: str) -> Any:
        return getattr(requests.Response, name, None)

    @property
    def content(self) -> Optional[bytes]:
        return super().content

    @content.setter
    def content(self, value: Union[str, bytes]) -> None:
        self._content = value.encode() if isinstance(value, str) else value



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#

from airbyte_cdk.sources.declarative.stream_slicers.stream_slicer import StreamSlicer
from airbyte_cdk.sources.declarative.stream_slicers.stream_slicer_test_read_decorator import (
    StreamSlicerTestReadDecorator,
)

__all__ = ["StreamSlicer", "StreamSlicerTestReadDecorator"]



#
# Copyright (c) 2025 Airbyte, Inc., all rights reserved.
#

import json
import logging
import pkgutil
from copy import deepcopy
from importlib import metadata
from types import ModuleType
from typing import Any, Dict, Iterator, List, Mapping, Optional, Set, Union

import orjson
import yaml
from jsonschema.exceptions import ValidationError
from jsonschema.validators import validate
from packaging.version import InvalidVersion, Version

from airbyte_cdk.config_observation import create_connector_config_control_message
from airbyte_cdk.connector_builder.models import (
    LogMessage as ConnectorBuilderLogMessage,
)
from airbyte_cdk.legacy.sources.declarative.declarative_source import DeclarativeSource
from airbyte_cdk.manifest_migrations.migration_handler import (
    ManifestMigrationHandler,
)
from airbyte_cdk.models import (
    AirbyteConnectionStatus,
    AirbyteMessage,
    AirbyteStateMessage,
    ConfiguredAirbyteCatalog,
    ConnectorSpecification,
    FailureType,
)
from airbyte_cdk.models.airbyte_protocol_serializers import AirbyteMessageSerializer
from airbyte_cdk.sources.declarative.checks import COMPONENTS_CHECKER_TYPE_MAPPING
from airbyte_cdk.sources.declarative.checks.connection_checker import ConnectionChecker
from airbyte_cdk.sources.declarative.interpolation import InterpolatedBoolean
from airbyte_cdk.sources.declarative.models.declarative_component_schema import (
    ConditionalStreams as ConditionalStreamsModel,
)
from airbyte_cdk.sources.declarative.models.declarative_component_schema import (
    DeclarativeStream as DeclarativeStreamModel,
)
from airbyte_cdk.sources.declarative.models.declarative_component_schema import (
    Spec as SpecModel,
)
from airbyte_cdk.sources.declarative.models.declarative_component_schema import (
    StateDelegatingStream as StateDelegatingStreamModel,
)
from airbyte_cdk.sources.declarative.parsers.custom_code_compiler import (
    get_registered_components_module,
)
from airbyte_cdk.sources.declarative.parsers.manifest_component_transformer import (
    ManifestComponentTransformer,
)
from airbyte_cdk.sources.declarative.parsers.manifest_normalizer import (
    ManifestNormalizer,
)
from airbyte_cdk.sources.declarative.parsers.manifest_reference_resolver import (
    ManifestReferenceResolver,
)
from airbyte_cdk.sources.declarative.parsers.model_to_component_factory import (
    ModelToComponentFactory,
)
from airbyte_cdk.sources.declarative.resolvers import COMPONENTS_RESOLVER_TYPE_MAPPING
from airbyte_cdk.sources.declarative.spec.spec import Spec
from airbyte_cdk.sources.message import MessageRepository
from airbyte_cdk.sources.streams.concurrent.abstract_stream import AbstractStream
from airbyte_cdk.sources.streams.core import Stream
from airbyte_cdk.sources.types import Config, ConnectionDefinition
from airbyte_cdk.sources.utils.slice_logger import (
    AlwaysLogSliceLogger,
    DebugSliceLogger,
    SliceLogger,
)
from airbyte_cdk.utils.traced_exception import AirbyteTracedException


def _get_declarative_component_schema() -> Dict[str, Any]:
    try:
        raw_component_schema = pkgutil.get_data(
            "airbyte_cdk", "sources/declarative/declarative_component_schema.yaml"
        )
        if raw_component_schema is not None:
            declarative_component_schema = yaml.load(raw_component_schema, Loader=yaml.SafeLoader)
            return declarative_component_schema  # type: ignore
        else:
            raise RuntimeError(
                "Failed to read manifest component json schema required for deduplication"
            )
    except FileNotFoundError as e:
        raise FileNotFoundError(
            f"Failed to read manifest component json schema required for deduplication: {e}"
        )


class ManifestDeclarativeSource(DeclarativeSource):
    """Declarative source defined by a manifest of low-code components that define source connector behavior"""

    def __init__(
        self,
        source_config: ConnectionDefinition,
        *,
        config: Mapping[str, Any] | None = None,
        debug: bool = False,
        emit_connector_builder_messages: bool = False,
        component_factory: Optional[ModelToComponentFactory] = None,
        migrate_manifest: Optional[bool] = False,
        normalize_manifest: Optional[bool] = False,
        config_path: Optional[str] = None,
    ) -> None:
        """
        Args:
            config: The provided config dict.
            source_config: The manifest of low-code components that describe the source connector.
            debug: True if debug mode is enabled.
            emit_connector_builder_messages: True if messages should be emitted to the connector builder.
            component_factory: optional factory if ModelToComponentFactory's default behavior needs to be tweaked.
            normalize_manifest: Optional flag to indicate if the manifest should be normalized.
            config_path: Optional path to the config file.
        """
        self.logger = logging.getLogger(f"airbyte.{self.name}")
        self._should_normalize = normalize_manifest
        self._should_migrate = migrate_manifest
        self._declarative_component_schema = _get_declarative_component_schema()
        # If custom components are needed, locate and/or register them.
        self.components_module: ModuleType | None = get_registered_components_module(config=config)
        # set additional attributes
        self._debug = debug
        self._emit_connector_builder_messages = emit_connector_builder_messages
        self._constructor = (
            component_factory
            if component_factory
            else ModelToComponentFactory(
                emit_connector_builder_messages=emit_connector_builder_messages,
                max_concurrent_async_job_count=source_config.get("max_concurrent_async_job_count"),
            )
        )
        self._message_repository = self._constructor.get_message_repository()
        self._slice_logger: SliceLogger = (
            AlwaysLogSliceLogger() if emit_connector_builder_messages else DebugSliceLogger()
        )

        # resolve all components in the manifest
        self._source_config = self._pre_process_manifest(dict(source_config))
        # validate resolved manifest against the declarative component schema
        self._validate_source()
        # apply additional post-processing to the manifest
        self._post_process_manifest()

        spec: Optional[Mapping[str, Any]] = self._source_config.get("spec")
        self._spec_component: Optional[Spec] = (
            self._constructor.create_component(SpecModel, spec, dict()) if spec else None
        )
        self._config = self._migrate_and_transform_config(config_path, config) or {}

    @property
    def resolved_manifest(self) -> Mapping[str, Any]:
        """
        Returns the resolved manifest configuration for the source.

        This property provides access to the internal source configuration as a mapping,
        which contains all settings and parameters required to define the source's behavior.

        Returns:
            Mapping[str, Any]: The resolved source configuration manifest.
        """
        return self._source_config

    def _pre_process_manifest(self, manifest: Dict[str, Any]) -> Dict[str, Any]:
        """
        Preprocesses the provided manifest dictionary by resolving any manifest references.

        This method modifies the input manifest in place, resolving references using the
        ManifestReferenceResolver to ensure all references within the manifest are properly handled.

        Args:
            manifest (Dict[str, Any]): The manifest dictionary to preprocess and resolve references in.

        Returns:
            None
        """
        # For ease of use we don't require the type to be specified at the top level manifest, but it should be included during processing
        manifest = self._fix_source_type(manifest)
        # Resolve references in the manifest
        resolved_manifest = ManifestReferenceResolver().preprocess_manifest(manifest)
        # Propagate types and parameters throughout the manifest
        propagated_manifest = ManifestComponentTransformer().propagate_types_and_parameters(
            "", resolved_manifest, {}
        )

        return propagated_manifest

    def _post_process_manifest(self) -> None:
        """
        Post-processes the manifest after validation.
        This method is responsible for any additional modifications or transformations needed
        after the manifest has been validated and before it is used in the source.
        """
        # apply manifest migration, if required
        self._migrate_manifest()
        # apply manifest normalization, if required
        self._normalize_manifest()

    def _normalize_manifest(self) -> None:
        """
        This method is used to normalize the manifest. It should be called after the manifest has been validated.

        Connector Builder UI rendering requires the manifest to be in a specific format.
         - references have been resolved
         - the commonly used definitions are extracted to the `definitions.linked.*`
        """
        if self._should_normalize:
            normalizer = ManifestNormalizer(self._source_config, self._declarative_component_schema)
            self._source_config = normalizer.normalize()

    def _migrate_and_transform_config(
        self,
        config_path: Optional[str],
        config: Optional[Config],
    ) -> Optional[Config]:
        if not config:
            return None
        if not self._spec_component:
            return config
        mutable_config = dict(config)
        self._spec_component.migrate_config(mutable_config)
        if mutable_config != config:
            if config_path:
                with open(config_path, "w") as f:
                    json.dump(mutable_config, f)
            self.message_repository.emit_message(
                create_connector_config_control_message(mutable_config)
            )
            # We have no mechanism for consuming the queue, so we print the messages to stdout
            for message in self.message_repository.consume_queue():
                print(orjson.dumps(AirbyteMessageSerializer.dump(message)).decode())
        self._spec_component.transform_config(mutable_config)
        return mutable_config

    def configure(self, config: Mapping[str, Any], temp_dir: str) -> Mapping[str, Any]:
        config = self._config or config
        return super().configure(config, temp_dir)

    def _migrate_manifest(self) -> None:
        """
        This method is used to migrate the manifest. It should be called after the manifest has been validated.
        The migration is done in place, so the original manifest is modified.

        The original manifest is returned if any error occurs during migration.
        """
        if self._should_migrate:
            manifest_migrator = ManifestMigrationHandler(self._source_config)
            self._source_config = manifest_migrator.apply_migrations()
            # validate migrated manifest against the declarative component schema
            self._validate_source()

    def _fix_source_type(self, manifest: Dict[str, Any]) -> Dict[str, Any]:
        """
        Fix the source type in the manifest. This is necessary because the source type is not always set in the manifest.
        """
        if "type" not in manifest:
            manifest["type"] = "DeclarativeSource"

        return manifest

    @property
    def message_repository(self) -> MessageRepository:
        return self._message_repository

    @property
    def dynamic_streams(self) -> List[Dict[str, Any]]:
        return self._dynamic_stream_configs(
            manifest=self._source_config,
            config=self._config,
            with_dynamic_stream_name=True,
        )

    def deprecation_warnings(self) -> List[ConnectorBuilderLogMessage]:
        return self._constructor.get_model_deprecations()

    @property
    def connection_checker(self) -> ConnectionChecker:
        check = self._source_config["check"]
        if "type" not in check:
            check["type"] = "CheckStream"
        check_stream = self._constructor.create_component(
            COMPONENTS_CHECKER_TYPE_MAPPING[check["type"]],
            check,
            dict(),
            emit_connector_builder_messages=self._emit_connector_builder_messages,
        )
        if isinstance(check_stream, ConnectionChecker):
            return check_stream
        else:
            raise ValueError(
                f"Expected to generate a ConnectionChecker component, but received {check_stream.__class__}"
            )

    def streams(self, config: Mapping[str, Any]) -> List[Union[Stream, AbstractStream]]:  # type: ignore  # we are migrating away from the AbstractSource and are expecting that this will only be called by ConcurrentDeclarativeSource or the Connector Builder
        """
        As a migration step, this method will return both legacy stream (Stream) and concurrent stream (AbstractStream).
        Once the migration is done, we can probably have this method throw "not implemented" as we figure out how to
        fully decouple this from the AbstractSource.
        """
        if self._spec_component:
            self._spec_component.validate_config(config)

        self._emit_manifest_debug_message(
            extra_args={
                "source_name": self.name,
                "parsed_config": json.dumps(self._source_config),
            }
        )

        stream_configs = (
            self._stream_configs(self._source_config, config=config) + self.dynamic_streams
        )

        api_budget_model = self._source_config.get("api_budget")
        if api_budget_model:
            self._constructor.set_api_budget(api_budget_model, config)

        source_streams = [
            self._constructor.create_component(
                (
                    StateDelegatingStreamModel
                    if stream_config.get("type") == StateDelegatingStreamModel.__name__
                    else DeclarativeStreamModel
                ),
                stream_config,
                config,
                emit_connector_builder_messages=self._emit_connector_builder_messages,
            )
            for stream_config in self._initialize_cache_for_parent_streams(deepcopy(stream_configs))
        ]
        return source_streams

    @staticmethod
    def _initialize_cache_for_parent_streams(
        stream_configs: List[Dict[str, Any]],
    ) -> List[Dict[str, Any]]:
        """Enable caching for parent streams unless explicitly disabled.

        Caching is enabled by default for parent streams to optimize performance when the same
        parent data is needed by multiple child streams. However, explicit `use_cache: false`
        settings are respected for streams that cannot use caching (e.g., scroll-based pagination
        APIs where caching causes duplicate records).
        """
        parent_streams = set()

        def _set_cache_if_not_disabled(requester: Dict[str, Any]) -> None:
            """Set use_cache to True only if not explicitly disabled."""
            if requester.get("use_cache") is not False:
                requester["use_cache"] = True

        def update_with_cache_parent_configs(
            parent_configs: list[dict[str, Any]],
        ) -> None:
            for parent_config in parent_configs:
                parent_streams.add(parent_config["stream"]["name"])
                if parent_config["stream"]["type"] == "StateDelegatingStream":
                    _set_cache_if_not_disabled(
                        parent_config["stream"]["full_refresh_stream"]["retriever"]["requester"]
                    )
                    _set_cache_if_not_disabled(
                        parent_config["stream"]["incremental_stream"]["retriever"]["requester"]
                    )
                else:
                    _set_cache_if_not_disabled(parent_config["stream"]["retriever"]["requester"])

        for stream_config in stream_configs:
            if stream_config.get("incremental_sync", {}).get("parent_stream"):
                parent_streams.add(stream_config["incremental_sync"]["parent_stream"]["name"])
                _set_cache_if_not_disabled(
                    stream_config["incremental_sync"]["parent_stream"]["retriever"]["requester"]
                )

            elif stream_config.get("retriever", {}).get("partition_router", {}):
                partition_router = stream_config["retriever"]["partition_router"]

                if isinstance(partition_router, dict) and partition_router.get(
                    "parent_stream_configs"
                ):
                    update_with_cache_parent_configs(partition_router["parent_stream_configs"])
                elif isinstance(partition_router, list):
                    for router in partition_router:
                        if router.get("parent_stream_configs"):
                            update_with_cache_parent_configs(router["parent_stream_configs"])

        for stream_config in stream_configs:
            if stream_config["name"] in parent_streams:
                if stream_config["type"] == "StateDelegatingStream":
                    _set_cache_if_not_disabled(
                        stream_config["full_refresh_stream"]["retriever"]["requester"]
                    )
                    _set_cache_if_not_disabled(
                        stream_config["incremental_stream"]["retriever"]["requester"]
                    )
                else:
                    _set_cache_if_not_disabled(stream_config["retriever"]["requester"])
        return stream_configs

    def spec(self, logger: logging.Logger) -> ConnectorSpecification:
        """
        Returns the connector specification (spec) as defined in the Airbyte Protocol. The spec is an object describing the possible
        configurations (e.g: username and password) which can be configured when running this connector. For low-code connectors, this
        will first attempt to load the spec from the manifest's spec block, otherwise it will load it from "spec.yaml" or "spec.json"
        in the project root.
        """
        self._configure_logger_level(logger)
        self._emit_manifest_debug_message(
            extra_args={
                "source_name": self.name,
                "parsed_config": json.dumps(self._source_config),
            }
        )

        return (
            self._spec_component.generate_spec() if self._spec_component else super().spec(logger)
        )

    def check(self, logger: logging.Logger, config: Mapping[str, Any]) -> AirbyteConnectionStatus:
        self._configure_logger_level(logger)
        return super().check(logger, config)

    def read(
        self,
        logger: logging.Logger,
        config: Mapping[str, Any],
        catalog: ConfiguredAirbyteCatalog,
        state: Optional[List[AirbyteStateMessage]] = None,
    ) -> Iterator[AirbyteMessage]:
        self._configure_logger_level(logger)
        yield from super().read(logger, config, catalog, state)

    def _configure_logger_level(self, logger: logging.Logger) -> None:
        """
        Set the log level to logging.DEBUG if debug mode is enabled
        """
        if self._debug:
            logger.setLevel(logging.DEBUG)

    def _validate_source(self) -> None:
        """
        Validates the connector manifest against the declarative component schema
        """

        try:
            validate(self._source_config, self._declarative_component_schema)
        except ValidationError as e:
            raise ValidationError(
                "Validation against json schema defined in declarative_component_schema.yaml schema failed"
            ) from e

        cdk_version_str = metadata.version("airbyte_cdk")
        cdk_version = self._parse_version(cdk_version_str, "airbyte-cdk")
        manifest_version_str = self._source_config.get("version")
        if manifest_version_str is None:
            raise RuntimeError(
                "Manifest version is not defined in the manifest. This is unexpected since it should be a required field. Please contact support."
            )
        manifest_version = self._parse_version(manifest_version_str, "manifest")

        if (cdk_version.major, cdk_version.minor, cdk_version.micro) == (0, 0, 0):
            # Skipping version compatibility check on unreleased dev branch
            pass
        elif (cdk_version.major, cdk_version.minor) < (
            manifest_version.major,
            manifest_version.minor,
        ):
            raise ValidationError(
                f"The manifest version {manifest_version!s} is greater than the airbyte-cdk package version ({cdk_version!s}). Your "
                f"manifest may contain features that are not in the current CDK version."
            )
        elif (manifest_version.major, manifest_version.minor) < (0, 29):
            raise ValidationError(
                f"The low-code framework was promoted to Beta in airbyte-cdk version 0.29.0 and contains many breaking changes to the "
                f"language. The manifest version {manifest_version!s} is incompatible with the airbyte-cdk package version "
                f"{cdk_version!s} which contains these breaking changes."
            )

    @staticmethod
    def _parse_version(
        version: str,
        version_type: str,
    ) -> Version:
        """Takes a semantic version represented as a string and splits it into a tuple.

        The fourth part (prerelease) is not returned in the tuple.

        Returns:
            Version: the parsed version object
        """
        try:
            parsed_version = Version(version)
        except InvalidVersion as ex:
            raise ValidationError(
                f"The {version_type} version '{version}' is not a valid version format."
            ) from ex
        else:
            # No exception
            return parsed_version

    def _stream_configs(
        self, manifest: Mapping[str, Any], config: Mapping[str, Any]
    ) -> List[Dict[str, Any]]:
        # This has a warning flag for static, but after we finish part 4 we'll replace manifest with self._source_config
        stream_configs = []
        for current_stream_config in manifest.get("streams", []):
            if (
                "type" in current_stream_config
                and current_stream_config["type"] == "ConditionalStreams"
            ):
                interpolated_boolean = InterpolatedBoolean(
                    condition=current_stream_config.get("condition"),
                    parameters={},
                )

                if interpolated_boolean.eval(config=config):
                    stream_configs.extend(current_stream_config.get("streams", []))
            else:
                if "type" not in current_stream_config:
                    current_stream_config["type"] = "DeclarativeStream"
                stream_configs.append(current_stream_config)
        return stream_configs

    def _dynamic_stream_configs(
        self,
        manifest: Mapping[str, Any],
        config: Mapping[str, Any],
        with_dynamic_stream_name: Optional[bool] = None,
    ) -> List[Dict[str, Any]]:
        dynamic_stream_definitions: List[Dict[str, Any]] = manifest.get("dynamic_streams", [])
        dynamic_stream_configs: List[Dict[str, Any]] = []
        seen_dynamic_streams: Set[str] = set()

        for dynamic_definition_index, dynamic_definition in enumerate(dynamic_stream_definitions):
            components_resolver_config = dynamic_definition["components_resolver"]

            if not components_resolver_config:
                raise ValueError(
                    f"Missing 'components_resolver' in dynamic definition: {dynamic_definition}"
                )

            resolver_type = components_resolver_config.get("type")
            if not resolver_type:
                raise ValueError(
                    f"Missing 'type' in components resolver configuration: {components_resolver_config}"
                )

            if resolver_type not in COMPONENTS_RESOLVER_TYPE_MAPPING:
                raise ValueError(
                    f"Invalid components resolver type '{resolver_type}'. "
                    f"Expected one of {list(COMPONENTS_RESOLVER_TYPE_MAPPING.keys())}."
                )

            if "retriever" in components_resolver_config:
                components_resolver_config["retriever"]["requester"]["use_cache"] = True

            # Create a resolver for dynamic components based on type
            if resolver_type == "HttpComponentsResolver":
                components_resolver = self._constructor.create_component(
                    model_type=COMPONENTS_RESOLVER_TYPE_MAPPING[resolver_type],
                    component_definition=components_resolver_config,
                    config=config,
                    stream_name=dynamic_definition.get("name"),
                )
            else:
                components_resolver = self._constructor.create_component(
                    model_type=COMPONENTS_RESOLVER_TYPE_MAPPING[resolver_type],
                    component_definition=components_resolver_config,
                    config=config,
                )

            stream_template_config = dynamic_definition["stream_template"]

            for dynamic_stream in components_resolver.resolve_components(
                stream_template_config=stream_template_config
            ):
                # Get the use_parent_parameters configuration from the dynamic definition
                # Default to True for backward compatibility, since connectors were already using it by default when this param was added
                use_parent_parameters = dynamic_definition.get("use_parent_parameters", True)

                dynamic_stream = {
                    **ManifestComponentTransformer().propagate_types_and_parameters(
                        "", dynamic_stream, {}, use_parent_parameters=use_parent_parameters
                    )
                }

                if "type" not in dynamic_stream:
                    dynamic_stream["type"] = "DeclarativeStream"

                # Ensure that each stream is created with a unique name
                name = dynamic_stream.get("name")

                if with_dynamic_stream_name:
                    dynamic_stream["dynamic_stream_name"] = dynamic_definition.get(
                        "name", f"dynamic_stream_{dynamic_definition_index}"
                    )

                if not isinstance(name, str):
                    raise ValueError(
                        f"Expected stream name {name} to be a string, got {type(name)}."
                    )

                if name in seen_dynamic_streams:
                    error_message = f"Dynamic streams list contains a duplicate name: {name}. Please contact Airbyte Support."
                    failure_type = FailureType.system_error

                    if resolver_type == "ConfigComponentsResolver":
                        error_message = f"Dynamic streams list contains a duplicate name: {name}. Please check your configuration."
                        failure_type = FailureType.config_error

                    raise AirbyteTracedException(
                        message=error_message,
                        internal_message=error_message,
                        failure_type=failure_type,
                    )

                seen_dynamic_streams.add(name)
                dynamic_stream_configs.append(dynamic_stream)

        return dynamic_stream_configs

    def _emit_manifest_debug_message(self, extra_args: dict[str, Any]) -> None:
        self.logger.debug("declarative source created from manifest", extra=extra_args)



#
# Copyright (c) 2025 Airbyte, Inc., all rights reserved.
#


from airbyte_cdk.manifest_migrations.manifest_migration import (
    TYPE_TAG,
    ManifestMigration,
    ManifestType,
)


class HttpRequesterUrlBaseToUrl(ManifestMigration):
    """
    This migration is responsible for migrating the `url_base` key to `url` in the HttpRequester component.
    The `url_base` key is expected to be a base URL, and the `url` key is expected to be a full URL.
    The migration will copy the value of `url_base` to `url`.
    """

    component_type = "HttpRequester"
    original_key = "url_base"
    replacement_key = "url"

    def should_migrate(self, manifest: ManifestType) -> bool:
        return manifest[TYPE_TAG] == self.component_type and self.original_key in list(
            manifest.keys()
        )

    def migrate(self, manifest: ManifestType) -> None:
        manifest[self.replacement_key] = manifest[self.original_key]
        manifest.pop(self.original_key, None)

    def validate(self, manifest: ManifestType) -> bool:
        """
        Validate the migration by checking if the `url` key is present and the `url_base` key is not.
        """
        return (
            self.replacement_key in manifest
            and self.original_key not in manifest
            and manifest[self.replacement_key] is not None
        )



# Copyright (c) 2025 Airbyte, Inc., all rights reserved.
"""Standalone CLI for the Airbyte CDK Manifest Server.

This CLI provides commands for running and managing the manifest server server.

**Installation:**

To use the manifest-server functionality, install the CDK with the manifest-server extra:

```bash
pip install airbyte-cdk[manifest-server]
# or
poetry install --extras manifest-server
```

**Usage:**

```bash
manifest-server start --port 8000
manifest-server info
manifest-server --help
```
"""

import rich_click as click

from ._info import info
from ._openapi import generate_openapi
from ._start import start


@click.group(
    help=__doc__.replace("\n", "\n\n"),  # Render docstring as help text (markdown)
    invoke_without_command=True,
)
@click.pass_context
def cli(
    ctx: click.Context,
) -> None:
    """Airbyte Manifest Server CLI."""

    if ctx.invoked_subcommand is None:
        # If no subcommand is provided, show the help message.
        click.echo(ctx.get_help())
        ctx.exit()


cli.add_command(start)
cli.add_command(info)
cli.add_command(generate_openapi)


def run() -> None:
    """Entry point for the manifest-server CLI."""
    cli()


if __name__ == "__main__":
    run()



"""
API Models for the Manifest Server Service.

This package contains all Pydantic models used for API requests and responses.
"""

from .dicts import ConnectorConfig, Manifest
from .manifest import (
    CheckRequest,
    CheckResponse,
    DiscoverRequest,
    DiscoverResponse,
    ErrorResponse,
    FullResolveRequest,
    ManifestResponse,
    RequestContext,
    ResolveRequest,
    StreamTestReadRequest,
)
from .stream import (
    AuxiliaryRequest,
    HttpRequest,
    HttpResponse,
    LogMessage,
    StreamReadPages,
    StreamReadResponse,
    StreamReadSlices,
)

__all__ = [
    # Typed Dicts
    "ConnectorConfig",
    "Manifest",
    # Manifest request/response models
    "RequestContext",
    "FullResolveRequest",
    "ManifestResponse",
    "StreamTestReadRequest",
    "ResolveRequest",
    "CheckRequest",
    "CheckResponse",
    "DiscoverRequest",
    "DiscoverResponse",
    "ErrorResponse",
    # Stream models
    "AuxiliaryRequest",
    "HttpRequest",
    "HttpResponse",
    "LogMessage",
    "StreamReadResponse",
    "StreamReadPages",
    "StreamReadSlices",
]



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#

import logging
import time
from abc import ABC, abstractmethod
from datetime import timezone
from enum import Enum
from io import IOBase
from os import makedirs, path
from typing import Any, Iterable, List, MutableMapping, Optional, Set, Tuple

from airbyte_protocol_dataclasses.models import FailureType
from wcmatch.glob import GLOBSTAR, globmatch

from airbyte_cdk.models import AirbyteRecordMessageFileReference
from airbyte_cdk.sources.file_based.config.abstract_file_based_spec import AbstractFileBasedSpec
from airbyte_cdk.sources.file_based.config.validate_config_transfer_modes import (
    include_identities_stream,
    preserve_directory_structure,
    use_file_transfer,
)
from airbyte_cdk.sources.file_based.exceptions import FileSizeLimitError
from airbyte_cdk.sources.file_based.file_record_data import FileRecordData
from airbyte_cdk.sources.file_based.remote_file import RemoteFile, UploadableRemoteFile
from airbyte_cdk.utils.datetime_helpers import ab_datetime_parse


class FileReadMode(Enum):
    READ = "r"
    READ_BINARY = "rb"


class AbstractFileBasedStreamReader(ABC):
    DATE_TIME_FORMAT = "%Y-%m-%dT%H:%M:%S.%fZ"
    FILE_RELATIVE_PATH = "file_relative_path"
    FILE_NAME = "file_name"
    LOCAL_FILE_PATH = "local_file_path"
    FILE_FOLDER = "file_folder"
    FILE_SIZE_LIMIT = 1_500_000_000

    def __init__(self) -> None:
        self._config = None

    @property
    def config(self) -> Optional[AbstractFileBasedSpec]:
        return self._config

    @config.setter
    @abstractmethod
    def config(self, value: AbstractFileBasedSpec) -> None:
        """
        FileBasedSource reads the config from disk and parses it, and once parsed, the source sets the config on its StreamReader.

        Note: FileBasedSource only requires the keys defined in the abstract config, whereas concrete implementations of StreamReader
        will require keys that (for example) allow it to authenticate with the 3rd party.

        Therefore, concrete implementations of AbstractFileBasedStreamReader's config setter should assert that `value` is of the correct
        config type for that type of StreamReader.
        """
        ...

    @abstractmethod
    def open_file(
        self, file: RemoteFile, mode: FileReadMode, encoding: Optional[str], logger: logging.Logger
    ) -> IOBase:
        """
        Return a file handle for reading.

        Many sources will be able to use smart_open to implement this method,
        for example:

        client = boto3.Session(...)
        return smart_open.open(remote_file.uri, transport_params={"client": client})
        """
        ...

    @abstractmethod
    def get_matching_files(
        self,
        globs: List[str],
        prefix: Optional[str],
        logger: logging.Logger,
    ) -> Iterable[RemoteFile]:
        """
        Return all files that match any of the globs.

        Example:

        The source has files "a.json", "foo/a.json", "foo/bar/a.json"

        If globs = ["*.json"] then this method returns ["a.json"].

        If globs = ["foo/*.json"] then this method returns ["foo/a.json"].

        Utility method `self.filter_files_by_globs` and `self.get_prefixes_from_globs`
        are available, which may be helpful when implementing this method.
        """
        ...

    def filter_files_by_globs_and_start_date(
        self, files: List[RemoteFile], globs: List[str]
    ) -> Iterable[RemoteFile]:
        """
        Utility method for filtering files based on globs.
        """
        start_date = (
            ab_datetime_parse(self.config.start_date)
            if self.config and self.config.start_date
            else None
        )
        seen = set()

        for file in files:
            if self.file_matches_globs(file, globs):
                last_modified = (
                    file.last_modified
                    if file.last_modified.tzinfo is not None
                    else file.last_modified.replace(tzinfo=timezone.utc)
                )
                if file.uri not in seen and (not start_date or last_modified >= start_date):
                    seen.add(file.uri)
                    yield file

    @staticmethod
    def file_matches_globs(file: RemoteFile, globs: List[str]) -> bool:
        # Use the GLOBSTAR flag to enable recursive ** matching
        # (https://facelessuser.github.io/wcmatch/wcmatch/#globstar)
        return any(globmatch(file.uri, g, flags=GLOBSTAR) for g in globs)

    @staticmethod
    def get_prefixes_from_globs(globs: List[str]) -> Set[str]:
        """
        Utility method for extracting prefixes from the globs.
        """
        prefixes = {glob.split("*")[0] for glob in globs}
        return set(filter(lambda x: bool(x), prefixes))

    def use_file_transfer(self) -> bool:
        if self.config:
            return use_file_transfer(self.config)
        return False

    def preserve_directory_structure(self) -> bool:
        # fall back to preserve subdirectories if config is not present or incomplete
        if self.config:
            return preserve_directory_structure(self.config)
        return True

    def include_identities_stream(self) -> bool:
        if self.config:
            return include_identities_stream(self.config)
        return False

    def upload(
        self, file: UploadableRemoteFile, local_directory: str, logger: logging.Logger
    ) -> Tuple[FileRecordData, AirbyteRecordMessageFileReference]:
        """
        This is required for connectors that will support writing to
        files. It will handle the logic to download,get,read,acquire or
        whatever is more efficient to get a file from the source.

        Args:
               file (RemoteFile): The remote file object containing URI and metadata.
               local_directory (str): The local directory path where the file will be downloaded.
               logger (logging.Logger): Logger for logging information and errors.

           Returns:
               AirbyteRecordMessageFileReference: A file reference object containing:
                   - staging_file_url (str): The absolute path to the referenced file in the staging area.
                   - file_size_bytes (int): The size of the referenced file in bytes.
                   - source_file_relative_path (str): The relative path to the referenced file in source.
        """
        if not isinstance(file, UploadableRemoteFile):
            raise TypeError(f"Expected UploadableRemoteFile, got {type(file)}")

        file_size = file.size

        if file_size > self.FILE_SIZE_LIMIT:
            message = (
                f"File size exceeds the {self.FILE_SIZE_LIMIT / 1e9} GB limit. File URI: {file.uri}"
            )
            raise FileSizeLimitError(
                message=message, internal_message=message, failure_type=FailureType.config_error
            )

        file_paths = self._get_file_transfer_paths(
            source_file_relative_path=file.source_file_relative_path,
            staging_directory=local_directory,
        )
        local_file_path = file_paths[self.LOCAL_FILE_PATH]
        file_relative_path = file_paths[self.FILE_RELATIVE_PATH]
        file_name = file_paths[self.FILE_NAME]

        logger.info(
            f"Starting to download the file {file.file_uri_for_logging} with size: {file_size / (1024 * 1024):,.2f} MB ({file_size / (1024 * 1024 * 1024):.2f} GB)"
        )
        start_download_time = time.time()

        file.download_to_local_directory(local_file_path)

        write_duration = time.time() - start_download_time
        logger.info(
            f"Finished downloading the file {file.file_uri_for_logging} and saved to {local_file_path} in {write_duration:,.2f} seconds."
        )

        file_record_data = FileRecordData(
            folder=file_paths[self.FILE_FOLDER],
            file_name=file_name,
            bytes=file_size,
            id=file.id,
            mime_type=file.mime_type,
            created_at=file.created_at,
            updated_at=file.updated_at,
            source_uri=file.source_uri,
        )
        file_reference = AirbyteRecordMessageFileReference(
            staging_file_url=local_file_path,
            source_file_relative_path=file_relative_path,
            file_size_bytes=file_size,
        )
        return file_record_data, file_reference

    def _get_file_transfer_paths(
        self, source_file_relative_path: str, staging_directory: str
    ) -> MutableMapping[str, Any]:
        """
        This method is used to get the file transfer paths for a given source file relative path and local directory.
        It returns a dictionary with the following keys:
            - FILE_RELATIVE_PATH: The relative path to file in reference to the staging directory.
            - LOCAL_FILE_PATH: The absolute path to the file.
            - FILE_NAME: The name of the referenced file.
            - FILE_FOLDER: The folder of the referenced file.
        """
        preserve_directory_structure = self.preserve_directory_structure()

        file_name = path.basename(source_file_relative_path)
        file_folder = path.dirname(source_file_relative_path)
        if preserve_directory_structure:
            # Remove left slashes from source path format to make relative path for writing locally
            file_relative_path = source_file_relative_path.lstrip("/")
        else:
            file_relative_path = file_name
        local_file_path = path.join(staging_directory, file_relative_path)
        # Ensure the local directory exists
        makedirs(path.dirname(local_file_path), exist_ok=True)

        file_paths = {
            self.FILE_RELATIVE_PATH: file_relative_path,
            self.LOCAL_FILE_PATH: local_file_path,
            self.FILE_NAME: file_name,
            self.FILE_FOLDER: file_folder,
        }
        return file_paths



# Copyright (c) 2025 Airbyte, Inc., all rights reserved.
"""Exceptions for the Airbyte CDK CLI."""

from dataclasses import dataclass


@dataclass(kw_only=True)
class ConnectorSecretWithNoValidVersionsError(Exception):
    """Error when a connector secret has no valid versions."""

    connector_name: str
    secret_name: str
    gcp_project_id: str

    def __str__(self) -> str:
        """Return a string representation of the exception."""
        from airbyte_cdk.cli.airbyte_cdk._secrets import _get_secret_url

        url = _get_secret_url(self.secret_name, self.gcp_project_id)
        return (
            f"No valid versions found for secret '{self.secret_name}' in connector '{self.connector_name}'. "
            f"Please check the following URL for more information:\n- {url}"
        )



# Copyright (c) 2025 Airbyte, Inc., all rights reserved.

import logging
from queue import Queue
from typing import Optional

from airbyte_cdk.sources.concurrent_source.stream_thread_exception import StreamThreadException
from airbyte_cdk.sources.message.repository import MessageRepository
from airbyte_cdk.sources.streams.concurrent.cursor import Cursor
from airbyte_cdk.sources.streams.concurrent.partitions.partition import Partition
from airbyte_cdk.sources.streams.concurrent.partitions.types import (
    PartitionCompleteSentinel,
    QueueItem,
)
from airbyte_cdk.sources.utils.slice_logger import SliceLogger


# Since moving all the connector builder workflow to the concurrent CDK which required correct ordering
# of grouping log messages onto the main write thread using the ConcurrentMessageRepository, this
# separate flow and class that was used to log slices onto this partition's message_repository
# should just be replaced by emitting messages directly onto the repository instead of an intermediary.
class PartitionLogger:
    """
    Helper class that provides a mechanism for passing a log message onto the current
    partitions message repository
    """

    def __init__(
        self,
        slice_logger: SliceLogger,
        logger: logging.Logger,
        message_repository: MessageRepository,
    ):
        self._slice_logger = slice_logger
        self._logger = logger
        self._message_repository = message_repository

    def log(self, partition: Partition) -> None:
        if self._slice_logger.should_log_slice_message(self._logger):
            self._message_repository.emit_message(
                self._slice_logger.create_slice_log_message(partition.to_slice())
            )


class PartitionReader:
    """
    Generates records from a partition and puts them in a queue.
    """

    _IS_SUCCESSFUL = True

    def __init__(
        self,
        queue: Queue[QueueItem],
        partition_logger: Optional[PartitionLogger] = None,
    ) -> None:
        """
        :param queue: The queue to put the records in.
        """
        self._queue = queue
        self._partition_logger = partition_logger

    def process_partition(self, partition: Partition, cursor: Cursor) -> None:
        """
        Process a partition and put the records in the output queue.
        When all the partitions are added to the queue, a sentinel is added to the queue to indicate that all the partitions have been generated.

        If an exception is encountered, the exception will be caught and put in the queue. This is very important because if we don't, the
        main thread will have no way to know that something when wrong and will wait until the timeout is reached

        This method is meant to be called from a thread.
        :param partition: The partition to read data from
        :return: None
        """
        try:
            if self._partition_logger:
                self._partition_logger.log(partition)

            for record in partition.read():
                self._queue.put(record)
                cursor.observe(record)
            cursor.close_partition(partition)
            self._queue.put(PartitionCompleteSentinel(partition, self._IS_SUCCESSFUL))
        except Exception as e:
            self._queue.put(StreamThreadException(e, partition.stream_name()))
            self._queue.put(PartitionCompleteSentinel(partition, not self._IS_SUCCESSFUL))



#
# Copyright (c) 2025 Airbyte, Inc., all rights reserved.
#


class ManifestMigrationException(Exception):
    """
    Raised when a migration error occurs in the manifest.
    """

    def __init__(self, message: str) -> None:
        super().__init__(f"Failed to migrate the manifest: {message}")



# Copyright (c) 2025 Airbyte, Inc., all rights reserved.
"""Info command for the manifest server CLI."""

from typing import Any, Optional

import rich_click as click

# Import server dependencies with graceful fallback
fastapi: Optional[Any] = None
uvicorn: Optional[Any] = None

try:
    import fastapi  # type: ignore[no-redef]
    import uvicorn  # type: ignore[no-redef]

    FASTAPI_AVAILABLE = True
except ImportError:
    FASTAPI_AVAILABLE = False


@click.command()
def info() -> None:
    """Show manifest server information and status."""
    if FASTAPI_AVAILABLE and fastapi is not None and uvicorn is not None:
        click.echo("✅ Manifest runner dependencies are installed")
        click.echo(f"   FastAPI version: {fastapi.__version__}")
        click.echo(f"   Uvicorn version: {uvicorn.__version__}")
    else:
        click.echo("❌ Manifest runner dependencies not installed")
        click.echo("   Install with: pip install airbyte-cdk[manifest-server]")



# Copyright (c) 2024 Airbyte, Inc., all rights reserved.

from dataclasses import dataclass
from enum import Enum
from typing import Optional, Union

import requests
from requests import HTTPError

from airbyte_cdk.models import FailureType
from airbyte_cdk.utils.airbyte_secrets_utils import filter_secrets


class ResponseAction(Enum):
    SUCCESS = "SUCCESS"
    RETRY = "RETRY"
    FAIL = "FAIL"
    IGNORE = "IGNORE"
    RESET_PAGINATION = "RESET_PAGINATION"
    RATE_LIMITED = "RATE_LIMITED"
    REFRESH_TOKEN_THEN_RETRY = "REFRESH_TOKEN_THEN_RETRY"


@dataclass
class ErrorResolution:
    response_action: Optional[ResponseAction] = None
    failure_type: Optional[FailureType] = None
    error_message: Optional[str] = None


def _format_exception_error_message(exception: Exception) -> str:
    return f"{type(exception).__name__}: {str(exception)}"


def _format_response_error_message(response: requests.Response) -> str:
    try:
        response.raise_for_status()
    except HTTPError as exception:
        return filter_secrets(
            f"Response was not ok: `{str(exception)}`. Response content is: {response.text}"
        )
    # We purposefully do not add the response.content because the response is "ok" so there might be sensitive information in the payload.
    # Feel free the
    return f"Unexpected response with HTTP status {response.status_code}"


def create_fallback_error_resolution(
    response_or_exception: Optional[Union[requests.Response, Exception]],
) -> ErrorResolution:
    if response_or_exception is None:
        # We do not expect this case to happen but if it does, it would be good to understand the cause and improve the error message
        error_message = "Error handler did not receive a valid response or exception. This is unexpected please contact Airbyte Support"
    elif isinstance(response_or_exception, Exception):
        error_message = _format_exception_error_message(response_or_exception)
    else:
        error_message = _format_response_error_message(response_or_exception)

    return ErrorResolution(
        response_action=ResponseAction.RETRY,
        failure_type=FailureType.system_error,
        error_message=error_message,
    )


SUCCESS_RESOLUTION = ErrorResolution(
    response_action=ResponseAction.SUCCESS, failure_type=None, error_message=None
)



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#

from dataclasses import InitVar, dataclass, field
from typing import Any, Dict, Mapping, Optional, Union

import requests

from airbyte_cdk.sources.declarative.decoders import (
    Decoder,
    JsonDecoder,
    PaginationDecoderDecorator,
)
from airbyte_cdk.sources.declarative.interpolation.interpolated_boolean import InterpolatedBoolean
from airbyte_cdk.sources.declarative.interpolation.interpolated_string import InterpolatedString
from airbyte_cdk.sources.declarative.requesters.paginators.strategies.pagination_strategy import (
    PaginationStrategy,
)
from airbyte_cdk.sources.types import Config, Record


@dataclass
class CursorPaginationStrategy(PaginationStrategy):
    """
    Pagination strategy that evaluates an interpolated string to define the next page token.

    Attributes:
        page_size (Optional[Union[str, int]]): the number of records to request
        cursor_value (Union[InterpolatedString, str]): template string evaluating to the cursor value
        config (Config): connection config
        stop_condition (Optional[InterpolatedBoolean]): template string evaluating when to stop paginating
        decoder (Decoder): decoder to decode the response
    """

    cursor_value: Union[InterpolatedString, str]
    config: Config
    parameters: InitVar[Mapping[str, Any]]
    page_size: Optional[Union[str, int]] = None
    stop_condition: Optional[Union[InterpolatedBoolean, str]] = None
    decoder: Decoder = field(
        default_factory=lambda: PaginationDecoderDecorator(decoder=JsonDecoder(parameters={}))
    )

    def __post_init__(self, parameters: Mapping[str, Any]) -> None:
        if isinstance(self.cursor_value, str):
            self._cursor_value = InterpolatedString.create(self.cursor_value, parameters=parameters)
        else:
            self._cursor_value = self.cursor_value
        if isinstance(self.stop_condition, str):
            self._stop_condition: Optional[InterpolatedBoolean] = InterpolatedBoolean(
                condition=self.stop_condition, parameters=parameters
            )
        else:
            self._stop_condition = self.stop_condition

        if isinstance(self.page_size, int) or (self.page_size is None):
            self._page_size = self.page_size
        else:
            page_size = InterpolatedString(self.page_size, parameters=parameters).eval(self.config)
            if not isinstance(page_size, int):
                raise Exception(f"{page_size} is of type {type(page_size)}. Expected {int}")
            self._page_size = page_size

    @property
    def initial_token(self) -> Optional[Any]:
        """
        CursorPaginationStrategy does not have an initial value because the next cursor is typically included
        in the response of the first request. For Resumable Full Refresh streams that checkpoint the page
        cursor, the next cursor should be read from the state or stream slice object.
        """
        return None

    def next_page_token(
        self,
        response: requests.Response,
        last_page_size: int,
        last_record: Optional[Record],
        last_page_token_value: Optional[Any] = None,
    ) -> Optional[Any]:
        decoded_response = next(self.decoder.decode(response))
        # The default way that link is presented in requests.Response is a string of various links (last, next, etc). This
        # is not indexable or useful for parsing the cursor, so we replace it with the link dictionary from response.links
        headers: Dict[str, Any] = dict(response.headers)
        headers["link"] = response.links
        if self._stop_condition:
            should_stop = self._stop_condition.eval(
                self.config,
                response=decoded_response,
                headers=headers,
                last_record=last_record,
                last_page_size=last_page_size,
            )
            if should_stop:
                return None
        token = self._cursor_value.eval(
            config=self.config,
            response=decoded_response,
            headers=headers,
            last_record=last_record,
            last_page_size=last_page_size,
        )
        return token if token else None

    def get_page_size(self) -> Optional[int]:
        return self._page_size



# Copyright (c) 2025 Airbyte, Inc., all rights reserved.
"""**Secret management commands.**

This module provides commands for managing secrets for Airbyte connectors.

**Usage:**

```bash
# Fetch secrets
airbyte-cdk secrets fetch --connector-name source-github
airbyte-cdk secrets fetch --connector-directory /path/to/connector
airbyte-cdk secrets fetch  # Run from within a connector directory

# List secrets (without fetching)
airbyte-cdk secrets list --connector-name source-github
airbyte-cdk secrets list --connector-directory /path/to/connector
```

**Usage without pre-installing (stateless):**

```bash
pipx run airbyte-cdk secrets fetch ...
uvx airbyte-cdk secrets fetch ...
```

The command retrieves secrets from Google Secret Manager based on connector
labels and writes them to the connector's `secrets` directory.
"""

from __future__ import annotations

import json
import logging
import os
from functools import lru_cache
from pathlib import Path
from typing import Any, cast

import requests
import rich_click as click
import yaml
from click import style
from rich.console import Console
from rich.table import Table

from airbyte_cdk.cli.airbyte_cdk.exceptions import ConnectorSecretWithNoValidVersionsError
from airbyte_cdk.utils.connector_paths import (
    resolve_connector_name,
    resolve_connector_name_and_directory,
)

GCP_PROJECT_ID: str = os.environ.get("GCP_PROJECT_ID", "") or "dataline-integration-testing"
# We put the `or` outside the `get()` because we want the `GCP_PROJECT_ID`
# env var to be ignored if it contains an empty string, such as in CI where the
# workflow might set it to a value that is itself actually missing or unset.
"""The GCP project ID to use for fetching integration test secrets."""

CONNECTOR_LABEL = "connector"
GLOBAL_MASK_KEYS_URL = "https://connectors.airbyte.com/files/registries/v0/specs_secrets_mask.yaml"

logger = logging.getLogger("airbyte-cdk.cli.secrets")

try:
    import google.auth.exceptions
    from google.cloud import secretmanager_v1 as secretmanager
    from google.cloud.secretmanager_v1 import Secret
except ImportError:
    # If the package is not installed, we will raise an error in the CLI command.
    google = None  # type: ignore
    secretmanager = None  # type: ignore
    Secret = None  # type: ignore


@click.group(
    name="secrets",
    help=__doc__.replace("\n", "\n\n"),  # Render docstring as help text (markdown) # type: ignore
)
def secrets_cli_group() -> None:
    """Secret management commands."""
    pass


@secrets_cli_group.command()
@click.argument(
    "connector",
    required=False,
    type=str,
    metavar="[CONNECTOR]",
)
@click.option(
    "--gcp-project-id",
    type=str,
    default=GCP_PROJECT_ID,
    help=(
        "GCP project ID for retrieving integration tests credentials. "
        "Defaults to the value of the `GCP_PROJECT_ID` environment variable, if set."
    ),
)
@click.option(
    "--print-ci-secrets-masks",
    help="Print GitHub CI mask for secrets.",
    type=bool,
    is_flag=True,
    default=None,
)
def fetch(
    connector: str | Path | None = None,
    gcp_project_id: str = GCP_PROJECT_ID,
    print_ci_secrets_masks: bool | None = None,
) -> None:
    """Fetch secrets for a connector from Google Secret Manager.

    This command fetches secrets for a connector from Google Secret Manager and writes them
    to the connector's secrets directory.

    [CONNECTOR] can be a connector name (e.g. 'source-pokeapi'), a path to a connector directory, or omitted to use the current working directory.
    If a string containing '/' is provided, it is treated as a path. Otherwise, it is treated as a connector name.

    If no connector name or directory is provided, we will look within the current working
    directory. If the current working directory is not a connector directory (e.g. starting
    with 'source-') and no connector name or path is provided, the process will fail.

    The `--print-ci-secrets-masks` option will print the GitHub CI mask for the secrets.
    This is useful for masking secrets in CI logs.

    WARNING: The `--print-ci-secrets-masks` option causes the secrets to be printed in clear text to
    `STDOUT`. For security reasons, this argument will be ignored if the `CI` environment
    variable is not set.
    """
    click.echo("Fetching secrets...", err=True)

    client = _get_gsm_secrets_client()
    connector_name, connector_directory = resolve_connector_name_and_directory(connector)
    secrets_dir = _get_secrets_dir(
        connector_directory=connector_directory,
        connector_name=connector_name,
        ensure_exists=True,
    )
    secrets = _fetch_secret_handles(
        connector_name=connector_name,
        gcp_project_id=gcp_project_id,
    )
    # Fetch and write secrets
    secret_count = 0
    exceptions = []

    for secret in secrets:
        secret_file_path = _get_secret_filepath(
            secrets_dir=secrets_dir,
            secret=secret,
        )
        try:
            _write_secret_file(
                secret=secret,
                client=client,
                file_path=secret_file_path,
                connector_name=connector_name,
                gcp_project_id=gcp_project_id,
            )
            click.echo(f"Secret written to: {secret_file_path.absolute()!s}", err=True)
            secret_count += 1
        except ConnectorSecretWithNoValidVersionsError as e:
            exceptions.append(e)
            click.echo(
                f"Failed to retrieve secret '{e.secret_name}': No enabled version found", err=True
            )

    if secret_count == 0 and not exceptions:
        click.echo(
            f"No secrets found for connector: '{connector_name}'",
            err=True,
        )

    if exceptions:
        error_message = f"Failed to retrieve {len(exceptions)} secret(s)"
        click.echo(
            style(
                error_message,
                fg="red",
            ),
            err=True,
        )
        if secret_count == 0:
            raise exceptions[0]

    if print_ci_secrets_masks and "CI" not in os.environ:
        click.echo(
            "The `--print-ci-secrets-masks` option is only available in CI environments. "
            "The `CI` env var is either not set or not set to a truthy value. "
            "Skipping printing secret masks.",
            err=True,
        )
        print_ci_secrets_masks = False
    elif print_ci_secrets_masks is None:
        # If not explicitly set, we check if we are in a CI environment
        # and set to True if so.
        print_ci_secrets_masks = os.environ.get("CI", "") != ""

    if print_ci_secrets_masks:
        _print_ci_secrets_masks(
            secrets_dir=secrets_dir,
        )


@secrets_cli_group.command("list")
@click.argument(
    "connector",
    required=False,
    type=str,
    metavar="[CONNECTOR]",
)
@click.option(
    "--gcp-project-id",
    type=str,
    default=GCP_PROJECT_ID,
    help=(
        "GCP project ID for retrieving integration tests credentials. "
        "Defaults to the value of the `GCP_PROJECT_ID` environment variable, if set."
    ),
)
def list_(
    connector: str | Path | None = None,
    *,
    gcp_project_id: str = GCP_PROJECT_ID,
) -> None:
    """List secrets for a connector from Google Secret Manager.

    This command fetches secrets for a connector from Google Secret Manager and prints
    them as a table.

    [CONNECTOR] can be a connector name (e.g. 'source-pokeapi'), a path to a connector directory, or omitted to use the current working directory.
    If a string containing '/' is provided, it is treated as a path. Otherwise, it is treated as a connector name.

    If no connector name or directory is provided, we will look within the current working
    directory. If the current working directory is not a connector directory (e.g. starting
    with 'source-') and no connector name or path is provided, the process will fail.
    """
    click.echo("Scanning secrets...", err=True)

    if connector and isinstance(connector, str) and "/" not in connector and "\\" not in connector:
        connector_name = connector
    else:
        connector_name, _ = resolve_connector_name_and_directory(connector)
    secrets: list[Secret] = _fetch_secret_handles(  # type: ignore
        connector_name=connector_name,
        gcp_project_id=gcp_project_id,
    )

    if not secrets:
        click.echo(
            f"No secrets found for connector: '{connector_name}'",
            err=True,
        )
        return
    # print a rich table with the secrets
    click.echo(
        style(
            f"Secrets for connector '{connector_name}' in project '{gcp_project_id}':",
            fg="green",
        )
    )

    console = Console()
    table = Table(title=f"'{connector_name}' Secrets")
    table.add_column("Name", justify="left", style="cyan", overflow="fold")
    table.add_column("Labels", justify="left", style="magenta", overflow="fold")
    table.add_column("Created", justify="left", style="blue", overflow="fold")
    for secret in secrets:
        full_secret_name = secret.name
        secret_name = _extract_secret_name(full_secret_name)
        secret_url = _get_secret_url(secret_name, gcp_project_id)
        table.add_row(
            f"[link={secret_url}]{secret_name}[/link]",
            "\n".join([f"{k}={v}" for k, v in secret.labels.items()]),
            str(secret.create_time),
        )

    console.print(table)


def _extract_secret_name(secret_name: str) -> str:
    """Extract the secret name from a fully qualified secret path.

    Handles different formats of secret names:
    - Full path: "projects/project-id/secrets/SECRET_NAME"
    - Already extracted: "SECRET_NAME"

    Args:
        secret_name: The secret name or path

    Returns:
        str: The extracted secret name without project prefix
    """
    if "/secrets/" in secret_name:
        return secret_name.split("/secrets/")[-1]
    return secret_name


def _get_secret_url(secret_name: str, gcp_project_id: str) -> str:
    """Generate a URL for a secret in the GCP Secret Manager console.

    Note: This URL itself does not contain secrets or sensitive information.
    The URL itself is only useful for valid logged-in users of the project, and it
    safe to print this URL in logs.

    Args:
        secret_name: The name of the secret in GCP.
        gcp_project_id: The GCP project ID.

    Returns:
        str: URL to the secret in the GCP console
    """
    # Ensure we have just the secret name without the project prefix
    secret_name = _extract_secret_name(secret_name)
    return f"https://console.cloud.google.com/security/secret-manager/secret/{secret_name}/versions?hl=en&project={gcp_project_id}"


def _fetch_secret_handles(
    connector_name: str,
    gcp_project_id: str = GCP_PROJECT_ID,
) -> list["Secret"]:  # type: ignore
    """Fetch secrets from Google Secret Manager."""
    if not secretmanager:
        raise ImportError(
            "google-cloud-secret-manager package is required for Secret Manager integration. "
            "Install it with 'pip install airbyte-cdk[dev]' "
            "or 'pip install google-cloud-secret-manager'."
        )

    client = _get_gsm_secrets_client()

    # List all secrets with the connector label
    parent = f"projects/{gcp_project_id}"
    filter_string = f"labels.{CONNECTOR_LABEL}={connector_name}"
    secrets = client.list_secrets(
        request=secretmanager.ListSecretsRequest(
            parent=parent,
            filter=filter_string,
        )
    )
    return [s for s in secrets]


def _write_secret_file(
    secret: "Secret",  # type: ignore
    client: "secretmanager.SecretManagerServiceClient",  # type: ignore
    file_path: Path,
    connector_name: str,
    gcp_project_id: str,
) -> None:
    """Write the most recent enabled version of a secret to a file.

    Lists all enabled versions of the secret and selects the most recent one.
    Raises ConnectorSecretWithNoValidVersionsError if no enabled versions are found.

    Args:
        secret: The secret to write to a file
        client: The Secret Manager client
        file_path: The path to write the secret to
        connector_name: The name of the connector
        gcp_project_id: The GCP project ID

    Raises:
        ConnectorSecretWithNoValidVersionsError: If no enabled version is found
    """
    # List all enabled versions of the secret.
    response = client.list_secret_versions(
        request={"parent": secret.name, "filter": "state:ENABLED"}
    )

    # The API returns versions pre-sorted in descending order, with the
    # 0th item being the latest version.
    versions = list(response)

    if not versions:
        secret_name = _extract_secret_name(secret.name)
        raise ConnectorSecretWithNoValidVersionsError(
            connector_name=connector_name,
            secret_name=secret_name,
            gcp_project_id=gcp_project_id,
        )

    enabled_version = versions[0]

    response = client.access_secret_version(name=enabled_version.name)
    file_path.write_text(response.payload.data.decode("UTF-8"))
    file_path.chmod(0o600)  # default to owner read/write only


def _get_secrets_dir(
    connector_directory: Path,
    connector_name: str,
    ensure_exists: bool = True,
) -> Path:
    _ = connector_name  # Unused, but it may be used in the future for logging
    secrets_dir = connector_directory / "secrets"
    if ensure_exists:
        secrets_dir.mkdir(parents=True, exist_ok=True)

        gitignore_path = secrets_dir / ".gitignore"
        if not gitignore_path.exists():
            gitignore_path.write_text("*")

    return secrets_dir


def _get_secret_filepath(
    secrets_dir: Path,
    secret: Secret,  # type: ignore
) -> Path:
    """Get the file path for a secret based on its labels."""
    if secret.labels and "filename" in secret.labels:
        return secrets_dir / f"{secret.labels['filename']}.json"

    return secrets_dir / "config.json"  # Default filename


def _get_gsm_secrets_client() -> "secretmanager.SecretManagerServiceClient":  # type: ignore
    """Get the Google Secret Manager client.

    If the `GCP_GSM_CREDENTIALS` environment variable is set, the client will be
    created using service account credentials from that JSON string. Otherwise, the
    client will fall back to Application Default Credentials (ADC), which supports
    user credentials from `gcloud auth application-default login`, GCE metadata
    server credentials, and other standard GCP authentication methods.
    """
    if not secretmanager:
        raise ImportError(
            "google-cloud-secret-manager package is required for Secret Manager integration. "
            "Install it with 'pip install airbyte-cdk[dev]' "
            "or 'pip install google-cloud-secret-manager'."
        )

    credentials_json = os.environ.get("GCP_GSM_CREDENTIALS")
    if credentials_json:
        click.echo(
            "Using GCP service account credentials from GCP_GSM_CREDENTIALS env var.", err=True
        )
        return cast(
            "secretmanager.SecretManagerServiceClient",
            secretmanager.SecretManagerServiceClient.from_service_account_info(
                json.loads(credentials_json)
            ),
        )

    click.echo(
        "GCP_GSM_CREDENTIALS not set. Using Application Default Credentials (ADC).", err=True
    )
    try:
        return secretmanager.SecretManagerServiceClient()
    except google.auth.exceptions.DefaultCredentialsError:
        raise ValueError(
            "No Google Cloud credentials found. "
            "Either set the `GCP_GSM_CREDENTIALS` environment variable with service account JSON, "
            "or run `gcloud auth application-default login` to authenticate with your user account."
        ) from None


def _print_ci_secrets_masks(
    secrets_dir: Path,
) -> None:
    """Print GitHub CI mask for secrets.

    https://docs.github.com/en/actions/writing-workflows/choosing-what-your-workflow-does/workflow-commands-for-github-actions#example-masking-an-environment-variable

    The env var `CI` is set to a truthy value in GitHub Actions, so we can use it to
    determine if we are in a CI environment. If not, we don't want to print the masks,
    as it will cause the secrets to be printed in clear text to STDOUT.
    """
    if not os.environ.get("CI", None):
        click.echo(
            "The `--print-ci-secrets-masks` option is only available in CI environments. "
            "The `CI` env var is either not set or not set to a truthy value. "
            "Skipping printing secret masks.",
            err=True,
        )
        return

    for secret_file_path in secrets_dir.glob("*.json"):
        config_dict = json.loads(secret_file_path.read_text())
        _print_ci_secrets_masks_for_config(config=config_dict)


def _print_ci_secret_mask_for_string(secret: str) -> None:
    """Print GitHub CI mask for a single secret string.

    We expect single-line secrets, but we also handle the case where the secret contains newlines.
    For multi-line secrets, we must print a secret mask for each line separately.
    """
    for line in secret.splitlines():
        if line.strip():  # Skip empty lines
            print(f"::add-mask::{line!s}")


def _print_ci_secret_mask_for_value(value: Any) -> None:
    """Print GitHub CI mask for a single secret value.

    Call this function for any values identified as secrets, regardless of type.
    """
    if isinstance(value, dict):
        # For nested dicts, we call recursively on each value
        for v in value.values():
            _print_ci_secret_mask_for_value(v)

        return

    if isinstance(value, list):
        # For lists, we call recursively on each list item
        for list_item in value:
            _print_ci_secret_mask_for_value(list_item)

        return

    # For any other types besides dict and list, we convert to string and mask each line
    # separately to handle multi-line secrets (e.g. private keys).
    for line in str(value).splitlines():
        if line.strip():  # Skip empty lines
            _print_ci_secret_mask_for_string(line)


def _print_ci_secrets_masks_for_config(
    config: dict[str, str] | list[Any] | Any,
) -> None:
    """Print GitHub CI mask for secrets config, navigating child nodes recursively."""
    if isinstance(config, list):
        # Check each item in the list to look for nested dicts that may contain secrets:
        for item in config:
            _print_ci_secrets_masks_for_config(item)

    elif isinstance(config, dict):
        for key, value in config.items():
            if _is_secret_property(key):
                logger.debug(f"Masking secret for config key: {key}")
                _print_ci_secret_mask_for_value(value)
            elif isinstance(value, (dict, list)):
                # Recursively check nested dicts and lists
                _print_ci_secrets_masks_for_config(value)


def _is_secret_property(property_name: str) -> bool:
    """Check if the property name is in the list of properties to mask.

    To avoid false negatives, we perform a case-insensitive check, and we include any property name
    that contains a rule entry, even if it is not an exact match.

    For example, if the rule entry is "password", we will also match "PASSWORD" and "my_password".
    """
    names_to_mask: list[str] = _get_spec_mask()
    if any([mask.lower() in property_name.lower() for mask in names_to_mask]):
        return True

    return False


@lru_cache
def _get_spec_mask() -> list[str]:
    """Get the list of properties to mask from the spec mask file."""
    response = requests.get(GLOBAL_MASK_KEYS_URL, allow_redirects=True)
    if not response.ok:
        logger.error(f"Failed to fetch spec mask: {response.content.decode('utf-8')}")
    try:
        return cast(list[str], yaml.safe_load(response.content)["properties"])
    except Exception as e:
        logger.error(f"Failed to parse spec mask: {e}")
        raise



# Copyright (c) 2024 Airbyte, Inc., all rights reserved.
"""Defines the `source-declarative-manifest` connector, which installs alongside CDK.

This file was originally imported from the dedicated connector directory, under the
`airbyte` monorepo.

Usage:

```
pipx install airbyte-cdk
source-declarative-manifest --help
source-declarative-manifest spec
...
```
"""

from __future__ import annotations

import argparse
import json
import pkgutil
import sys
import traceback
from collections.abc import MutableMapping
from pathlib import Path
from typing import Any, cast

import orjson
import yaml

from airbyte_cdk.entrypoint import AirbyteEntrypoint, launch
from airbyte_cdk.models import (
    AirbyteErrorTraceMessage,
    AirbyteMessage,
    AirbyteMessageSerializer,
    AirbyteStateMessage,
    AirbyteTraceMessage,
    ConfiguredAirbyteCatalog,
    ConnectorSpecificationSerializer,
    TraceType,
    Type,
)
from airbyte_cdk.sources.declarative.concurrent_declarative_source import (
    ConcurrentDeclarativeSource,
)
from airbyte_cdk.sources.declarative.yaml_declarative_source import YamlDeclarativeSource
from airbyte_cdk.sources.source import TState
from airbyte_cdk.utils.datetime_helpers import ab_datetime_now


class SourceLocalYaml(YamlDeclarativeSource):
    """
    Declarative source defined by a yaml file in the local filesystem
    """

    def __init__(
        self,
        catalog: ConfiguredAirbyteCatalog | None,
        config: MutableMapping[str, Any] | None,
        state: TState,
        config_path: str | None = None,
        **kwargs: Any,
    ) -> None:
        """
        HACK!
            Problem: YamlDeclarativeSource relies on the calling module name/path to find the yaml file.
            Implication: If you call YamlDeclarativeSource directly it will look for the yaml file in the wrong place. (e.g. the airbyte-cdk package)
            Solution: Subclass YamlDeclarativeSource from the same location as the manifest to load.

            When can we remove this?
                When the airbyte-cdk is updated to not rely on the calling module name/path to find the yaml file.
                When all manifest connectors are updated to use the new airbyte-cdk.
                When all manifest connectors are updated to use the source-declarative-manifest as the base image.
        """
        super().__init__(
            catalog=catalog,
            config=config,
            state=state,  # type: ignore [arg-type]
            path_to_yaml="manifest.yaml",
            config_path=config_path,
        )


def _is_local_manifest_command(args: list[str]) -> bool:
    # Check for a local manifest.yaml file
    return Path("/airbyte/integration_code/source_declarative_manifest/manifest.yaml").exists()


def handle_command(args: list[str]) -> None:
    if _is_local_manifest_command(args):
        handle_local_manifest_command(args)
    else:
        handle_remote_manifest_command(args)


def _get_local_yaml_source(args: list[str]) -> SourceLocalYaml:
    try:
        parsed_args = AirbyteEntrypoint.parse_args(args)
        config, catalog, state = _parse_inputs_into_config_catalog_state(parsed_args)
        return SourceLocalYaml(
            config=config,
            catalog=catalog,
            state=state,
            config_path=parsed_args.config if hasattr(parsed_args, "config") else None,
        )
    except Exception as error:
        print(
            orjson.dumps(
                AirbyteMessageSerializer.dump(
                    AirbyteMessage(
                        type=Type.TRACE,
                        trace=AirbyteTraceMessage(
                            type=TraceType.ERROR,
                            emitted_at=ab_datetime_now().to_epoch_millis(),
                            error=AirbyteErrorTraceMessage(
                                message=f"Error starting the sync. This could be due to an invalid configuration or catalog. Please contact Support for assistance. Error: {error}",
                                stack_trace=traceback.format_exc(),
                            ),
                        ),
                    )
                )
            ).decode()
        )
        raise error


def handle_local_manifest_command(args: list[str]) -> None:
    source = _get_local_yaml_source(args)
    launch(
        source=source,
        args=args,
    )


def handle_remote_manifest_command(args: list[str]) -> None:
    """Overrides the spec command to return the generalized spec for the declarative manifest source.

    This is different from a typical low-code, but built and published separately source built as a ManifestDeclarativeSource,
    because that will have a spec method that returns the spec for that specific source. Other than spec,
    the generalized connector behaves the same as any other, since the manifest is provided in the config.
    """
    if args[0] == "spec":
        json_spec = pkgutil.get_data(
            "airbyte_cdk.cli.source_declarative_manifest",
            "spec.json",
        )
        if json_spec is None:
            raise FileNotFoundError(
                "Could not find `spec.json` file for source-declarative-manifest"
            )

        spec_obj = json.loads(json_spec)
        spec = ConnectorSpecificationSerializer.load(spec_obj)

        message = AirbyteMessage(type=Type.SPEC, spec=spec)
        print(AirbyteEntrypoint.airbyte_message_to_string(message))
    else:
        source = create_declarative_source(args)
        launch(
            source=source,
            args=args,
        )


def create_declarative_source(
    args: list[str],
) -> ConcurrentDeclarativeSource:  # type: ignore [type-arg]
    """Creates the source with the injected config.

    This essentially does what other low-code sources do at build time, but at runtime,
    with a user-provided manifest in the config. This better reflects what happens in the
    connector builder.
    """
    try:
        config: MutableMapping[str, Any] | None
        catalog: ConfiguredAirbyteCatalog | None
        state: list[AirbyteStateMessage]

        parsed_args = AirbyteEntrypoint.parse_args(args)
        config, catalog, state = _parse_inputs_into_config_catalog_state(parsed_args)

        if config is None:
            raise ValueError(
                "Invalid config: `__injected_declarative_manifest` should be provided at the root "
                "of the config or using the --manifest-path argument."
            )

        # If a manifest_path is provided in the args, inject it into the config
        if hasattr(parsed_args, "manifest_path") and parsed_args.manifest_path:
            injected_manifest = _parse_manifest_from_file(parsed_args.manifest_path)
            if injected_manifest:
                config["__injected_declarative_manifest"] = injected_manifest

        if "__injected_declarative_manifest" not in config:
            raise ValueError(
                "Invalid config: `__injected_declarative_manifest` should be provided at the root "
                "of the config or using the --manifest-path argument. "
                f"Config only has keys: {list(config.keys() if config else [])}"
            )
        if not isinstance(config["__injected_declarative_manifest"], dict):
            raise ValueError(
                "Invalid config: `__injected_declarative_manifest` should be a dictionary, "
                f"but got type: {type(config['__injected_declarative_manifest'])}"
            )

        if hasattr(parsed_args, "components_path") and parsed_args.components_path:
            _register_components_from_file(parsed_args.components_path)

        return ConcurrentDeclarativeSource(
            config=config,
            catalog=catalog,
            state=state,
            source_config=cast(dict[str, Any], config["__injected_declarative_manifest"]),
            config_path=parsed_args.config if hasattr(parsed_args, "config") else None,
        )
    except Exception as error:
        print(
            orjson.dumps(
                AirbyteMessageSerializer.dump(
                    AirbyteMessage(
                        type=Type.TRACE,
                        trace=AirbyteTraceMessage(
                            type=TraceType.ERROR,
                            emitted_at=ab_datetime_now().to_epoch_millis(),
                            error=AirbyteErrorTraceMessage(
                                message=f"Error starting the sync. This could be due to an invalid configuration or catalog. Please contact Support for assistance. Error: {error}",
                                stack_trace=traceback.format_exc(),
                            ),
                        ),
                    )
                )
            ).decode()
        )
        raise error


def _parse_inputs_into_config_catalog_state(
    parsed_args: argparse.Namespace,
) -> tuple[
    MutableMapping[str, Any] | None,
    ConfiguredAirbyteCatalog | None,
    list[AirbyteStateMessage],
]:
    config = (
        ConcurrentDeclarativeSource.read_config(parsed_args.config)
        if hasattr(parsed_args, "config")
        else None
    )
    catalog = (
        ConcurrentDeclarativeSource.read_catalog(parsed_args.catalog)
        if hasattr(parsed_args, "catalog")
        else None
    )
    state = (
        ConcurrentDeclarativeSource.read_state(parsed_args.state)
        if hasattr(parsed_args, "state")
        else []
    )

    return config, catalog, state


def _parse_manifest_from_file(filepath: str) -> dict[str, Any] | None:
    """Extract and parse a manifest file specified in the args."""
    try:
        with open(filepath, "r", encoding="utf-8") as manifest_file:
            manifest_content = yaml.safe_load(manifest_file)
            if manifest_content is None:
                raise ValueError(f"Manifest file at {filepath} is empty")
            if not isinstance(manifest_content, dict):
                raise ValueError(f"Manifest must be a dictionary, got {type(manifest_content)}")
            return manifest_content
    except Exception as error:
        raise ValueError(f"Failed to load manifest file from {filepath}: {error}")


def _register_components_from_file(filepath: str) -> None:
    """Load and register components from a Python file specified in the args."""
    import importlib.util
    import sys

    components_path = Path(filepath)

    module_name = "components"
    sdm_module_name = "source_declarative_manifest.components"

    # Create module spec
    spec = importlib.util.spec_from_file_location(module_name, components_path)
    if spec is None or spec.loader is None:
        raise ImportError(f"Could not load module from {components_path}")

    # Create module and execute code, registering the module before executing its code
    # To avoid issues with dataclasses that look up the module
    module = importlib.util.module_from_spec(spec)
    sys.modules[module_name] = module
    sys.modules[sdm_module_name] = module

    spec.loader.exec_module(module)


def run() -> None:
    """Run the `source-declarative-manifest` CLI.

    Args are detected from the command line, and the appropriate command is executed.
    """
    args: list[str] = sys.argv[1:]
    handle_command(args)



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#

import json
import re
from typing import Any

from referencing import Registry, Resource
from referencing.jsonschema import DRAFT7


def resolve_refs(schema: dict[str, Any]) -> dict[str, Any]:
    """
    For spec schemas generated using Pydantic models, the resulting JSON schema can contain refs between object
    relationships.
    """
    resource = Resource.from_contents(schema, default_specification=DRAFT7)
    registry = Registry().with_resource("", resource)
    resolver = registry.resolver()
    str_schema = json.dumps(schema)
    for ref_block in re.findall(r'{"\$ref": "#\/definitions\/.+?(?="})"}', str_schema):
        ref = json.loads(ref_block)["$ref"]
        resolved = resolver.lookup(ref).contents
        str_schema = str_schema.replace(ref_block, json.dumps(resolved))
    pyschema: dict[str, Any] = json.loads(str_schema)
    del pyschema["definitions"]
    return pyschema



#
# Copyright (c) 2025 Airbyte, Inc., all rights reserved.
#

import json
import logging
import uuid
from dataclasses import InitVar, dataclass, field
from pathlib import Path
from typing import Any, Mapping, Optional, Union

from airbyte_cdk.models import AirbyteRecordMessageFileReference
from airbyte_cdk.sources.declarative.extractors.record_extractor import RecordExtractor
from airbyte_cdk.sources.declarative.interpolation.interpolated_string import (
    InterpolatedString,
)
from airbyte_cdk.sources.declarative.partition_routers.substream_partition_router import (
    SafeResponse,
)
from airbyte_cdk.sources.declarative.requesters import Requester
from airbyte_cdk.sources.declarative.types import Record, StreamSlice
from airbyte_cdk.sources.types import Config
from airbyte_cdk.sources.utils.files_directory import get_files_directory

from .file_uploader import FileUploader
from .file_writer import FileWriter

logger = logging.getLogger("airbyte")


@dataclass
class DefaultFileUploader(FileUploader):
    """
    File uploader class
    Handles the upload logic: fetching the download target, making the request via its requester, determining the file path, and calling self.file_writer.write()
    Different types of file_writer:BaseFileWriter can be injected to handle different file writing strategies.
    """

    requester: Requester
    download_target_extractor: RecordExtractor
    config: Config
    file_writer: FileWriter
    parameters: InitVar[Mapping[str, Any]]

    filename_extractor: Optional[Union[InterpolatedString, str]] = None

    def __post_init__(self, parameters: Mapping[str, Any]) -> None:
        if self.filename_extractor:
            self.filename_extractor = InterpolatedString.create(
                self.filename_extractor,
                parameters=parameters,
            )

    def upload(self, record: Record) -> None:
        mocked_response = SafeResponse()
        mocked_response.content = json.dumps(record.data).encode()
        download_targets = list(self.download_target_extractor.extract_records(mocked_response))
        if not download_targets:
            raise ValueError("No download targets found")

        download_target = download_targets[0]  # we just expect one download target
        if not isinstance(download_target, str):
            raise ValueError(
                f"download_target is expected to be a str but was {type(download_target)}: {download_target}"
            )

        response = self.requester.send_request(
            stream_slice=StreamSlice(
                partition={}, cursor_slice={}, extra_fields={"download_target": download_target}
            ),
        )

        files_directory = Path(get_files_directory())

        file_name = (
            self.filename_extractor.eval(self.config, record=record)
            if self.filename_extractor
            else str(uuid.uuid4())
        )
        file_name = file_name.lstrip("/")
        file_relative_path = Path(record.stream_name) / Path(file_name)

        full_path = files_directory / file_relative_path
        full_path.parent.mkdir(parents=True, exist_ok=True)

        file_size_bytes = self.file_writer.write(full_path, content=response.content)

        logger.info("File uploaded successfully")
        logger.info(f"File url: {str(full_path)}")
        logger.info(f"File size: {file_size_bytes / 1024} KB")
        logger.info(f"File relative path: {str(file_relative_path)}")

        record.file_reference = AirbyteRecordMessageFileReference(
            staging_file_url=str(full_path),
            source_file_relative_path=str(file_relative_path),
            file_size_bytes=file_size_bytes,
        )



#
# Copyright (c) 2024 Airbyte, Inc., all rights reserved.
#
import logging
from typing import Iterable, Tuple

from airbyte_cdk.models import AirbyteRecordMessageFileReference
from airbyte_cdk.sources.file_based.file_based_stream_reader import AbstractFileBasedStreamReader
from airbyte_cdk.sources.file_based.file_record_data import FileRecordData
from airbyte_cdk.sources.file_based.remote_file import UploadableRemoteFile
from airbyte_cdk.sources.utils.files_directory import get_files_directory


class FileTransfer:
    def __init__(self) -> None:
        self._local_directory = get_files_directory()

    def upload(
        self,
        file: UploadableRemoteFile,
        stream_reader: AbstractFileBasedStreamReader,
        logger: logging.Logger,
    ) -> Iterable[Tuple[FileRecordData, AirbyteRecordMessageFileReference]]:
        try:
            yield stream_reader.upload(
                file=file, local_directory=self._local_directory, logger=logger
            )
        except Exception as ex:
            logger.error("An error has occurred while getting file: %s", str(ex))
            raise ex



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#

import csv
import gzip
import io
import json
import logging
from dataclasses import dataclass
from io import BufferedIOBase, TextIOWrapper
from typing import Any, List, Optional

import orjson
import requests

from airbyte_cdk.models import FailureType
from airbyte_cdk.sources.declarative.decoders.decoder import DECODER_OUTPUT_TYPE, Decoder
from airbyte_cdk.sources.declarative.decoders.decoder_parser import (
    PARSER_OUTPUT_TYPE,
    PARSERS_BY_HEADER_TYPE,
    PARSERS_TYPE,
    Parser,
)
from airbyte_cdk.utils import AirbyteTracedException

logger = logging.getLogger("airbyte")


@dataclass
class GzipParser(Parser):
    inner_parser: Parser

    def parse(self, data: BufferedIOBase) -> PARSER_OUTPUT_TYPE:
        """
        Decompress gzipped bytes and pass decompressed data to the inner parser.

        IMPORTANT:
            - If the data is not gzipped, reset the pointer and pass the data to the inner parser as is.

        Note:
            - The data is not decoded by default.
        """

        with gzip.GzipFile(fileobj=data, mode="rb") as gzipobj:
            yield from self.inner_parser.parse(gzipobj)


@dataclass
class JsonParser(Parser):
    encoding: str = "utf-8"

    def parse(self, data: BufferedIOBase) -> PARSER_OUTPUT_TYPE:
        """
        Attempts to deserialize data using orjson library. As an extra layer of safety we fallback on the json library to deserialize the data.
        """
        raw_data = data.read()
        body_json = self._parse_orjson(raw_data) or self._parse_json(raw_data)

        if body_json is None:
            raise AirbyteTracedException(
                message="Response JSON data failed to be parsed. See logs for more information.",
                internal_message=f"Response JSON data failed to be parsed.",
                failure_type=FailureType.system_error,
            )

        if isinstance(body_json, list):
            yield from body_json
        else:
            yield from [body_json]

    def _parse_orjson(self, raw_data: bytes) -> Optional[Any]:
        try:
            return orjson.loads(raw_data.decode(self.encoding))
        except Exception as exc:
            logger.debug(
                f"Failed to parse JSON data using orjson library. Falling back to json library. {exc}"
            )
            return None

    def _parse_json(self, raw_data: bytes) -> Optional[Any]:
        try:
            return json.loads(raw_data.decode(self.encoding))
        except Exception as exc:
            logger.error(f"Failed to parse JSON data using json library. {exc}")
            return None


@dataclass
class JsonLineParser(Parser):
    encoding: Optional[str] = "utf-8"

    def parse(self, data: BufferedIOBase) -> PARSER_OUTPUT_TYPE:
        for line in data:
            try:
                yield json.loads(line.decode(encoding=self.encoding or "utf-8"))
            except json.JSONDecodeError as e:
                logger.warning(f"Cannot decode/parse line {line!r} as JSON, error: {e}")


@dataclass
class CsvParser(Parser):
    # TODO: migrate implementation to re-use file-base classes
    encoding: Optional[str] = "utf-8"
    delimiter: Optional[str] = ","
    set_values_to_none: Optional[List[str]] = None

    def _get_delimiter(self) -> Optional[str]:
        """
        Get delimiter from the configuration. Check for the escape character and decode it.
        """
        if self.delimiter is not None:
            if self.delimiter.startswith("\\"):
                self.delimiter = self.delimiter.encode("utf-8").decode("unicode_escape")

        return self.delimiter

    def parse(self, data: BufferedIOBase) -> PARSER_OUTPUT_TYPE:
        """
        Parse CSV data from decompressed bytes.
        """
        text_data = TextIOWrapper(data, encoding=self.encoding)  # type: ignore
        reader = csv.DictReader(text_data, delimiter=self._get_delimiter() or ",")
        for row in reader:
            if self.set_values_to_none:
                row = {k: (None if v in self.set_values_to_none else v) for k, v in row.items()}
            yield row


class CompositeRawDecoder(Decoder):
    """
    Decoder strategy to transform a requests.Response into a PARSER_OUTPUT_TYPE
    passed response.raw to parser(s).

    Note: response.raw is not decoded/decompressed by default. Parsers should be instantiated recursively.

    Example:
        composite_raw_decoder = CompositeRawDecoder(
            parser=GzipParser(
                inner_parser=JsonLineParser(encoding="iso-8859-1")
            )
        )
    """

    def __init__(
        self,
        parser: Parser,
        stream_response: bool = True,
        parsers_by_header: PARSERS_BY_HEADER_TYPE = None,
    ) -> None:
        # since we moved from using `dataclass` to `__init__` method,
        # we need to keep using the `parser` to be able to resolve the depenencies
        # between the parsers correctly.
        self.parser = parser

        self._parsers_by_header = parsers_by_header if parsers_by_header else {}
        self._stream_response = stream_response

    @classmethod
    def by_headers(
        cls,
        parsers: PARSERS_TYPE,
        stream_response: bool,
        fallback_parser: Parser,
    ) -> "CompositeRawDecoder":
        """
        Create a CompositeRawDecoder instance based on header values.

        Args:
            parsers (PARSERS_TYPE): A list of tuples where each tuple contains headers, header values, and a parser.
            stream_response (bool): A flag indicating whether the response should be streamed.
            fallback_parser (Parser): A parser to use if no matching header is found.

        Returns:
            CompositeRawDecoder: An instance of CompositeRawDecoder configured with the provided parsers.
        """
        parsers_by_header = {}
        for headers, header_values, parser in parsers:
            for header in headers:
                parsers_by_header[header] = {header_value: parser for header_value in header_values}
        return cls(fallback_parser, stream_response, parsers_by_header)

    def is_stream_response(self) -> bool:
        return self._stream_response

    def decode(self, response: requests.Response) -> DECODER_OUTPUT_TYPE:
        parser = self._select_parser(response)
        if self.is_stream_response():
            # urllib mentions that some interfaces don't play nice with auto_close
            # More info here: https://urllib3.readthedocs.io/en/stable/user-guide.html#using-io-wrappers-with-response-content
            # We have indeed observed some issues with CSV parsing.
            # Hence, we will manage the closing of the file ourselves until we find a better solution.
            response.raw.auto_close = False
            yield from parser.parse(
                data=response.raw,  # type: ignore[arg-type]
            )
            response.raw.close()
        else:
            yield from parser.parse(data=io.BytesIO(response.content))

    def _select_parser(self, response: requests.Response) -> Parser:
        """
        Selects the appropriate parser based on the response headers.

        This method iterates through the `_parsers_by_header` dictionary to find a matching parser
        based on the headers in the response. If a matching header and header value are found,
        the corresponding parser is returned. If no match is found, the default parser is returned.

        Args:
            response (requests.Response): The HTTP response object containing headers to check.

        Returns:
            Parser: The parser corresponding to the matched header value, or the default parser if no match is found.
        """
        for header, parser_by_header_value in self._parsers_by_header.items():
            if (
                header in response.headers
                and response.headers[header] in parser_by_header_value.keys()
            ):
                return parser_by_header_value[response.headers[header]]
        return self.parser



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#

from typing import Any, Union

from airbyte_cdk.models import AirbyteMessage
from airbyte_cdk.sources.concurrent_source.partition_generation_completed_sentinel import (
    PartitionGenerationCompletedSentinel,
)
from airbyte_cdk.sources.streams.concurrent.partitions.partition import Partition
from airbyte_cdk.sources.types import Record


class PartitionCompleteSentinel:
    """
    A sentinel object indicating all records for a partition were produced.
    Includes a pointer to the partition that was processed.
    """

    def __init__(self, partition: Partition, is_successful: bool = True):
        """
        :param partition: The partition that was processed
        """
        self.partition = partition
        self.is_successful = is_successful

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, PartitionCompleteSentinel):
            return self.partition == other.partition
        return False


"""
Typedef representing the items that can be added to the ThreadBasedConcurrentStream
"""
QueueItem = Union[
    Record,
    Partition,
    PartitionCompleteSentinel,
    PartitionGenerationCompletedSentinel,
    Exception,
    AirbyteMessage,
]



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#

from typing import Any


class ExceptionWithDisplayMessage(Exception):
    """
    Exception that can be used to display a custom message to the user.
    """

    def __init__(self, display_message: str, **kwargs: Any):
        super().__init__(**kwargs)
        self.display_message = display_message

    def __str__(self) -> str:
        return f'ExceptionWithDisplayMessage: "{self.display_message}"'



#
# Copyright (c) 2025 Airbyte, Inc., all rights reserved.
#

from dataclasses import InitVar, dataclass, field
from typing import Any, List, Mapping, MutableMapping, Optional, Union

from airbyte_cdk.sources.declarative.interpolation.interpolated_nested_mapping import NestedMapping
from airbyte_cdk.sources.declarative.models.declarative_component_schema import (
    RequestBodyGraphQL,
    RequestBodyJsonObject,
    RequestBodyPlainText,
    RequestBodyUrlEncodedForm,
)
from airbyte_cdk.sources.declarative.requesters.request_options.interpolated_nested_request_input_provider import (
    InterpolatedNestedRequestInputProvider,
)
from airbyte_cdk.sources.declarative.requesters.request_options.interpolated_request_input_provider import (
    InterpolatedRequestInputProvider,
)
from airbyte_cdk.sources.declarative.requesters.request_options.request_options_provider import (
    RequestOptionsProvider,
)
from airbyte_cdk.sources.types import Config, StreamSlice, StreamState

RequestInput = Union[str, Mapping[str, str]]
ValidRequestTypes = (str, list)


@dataclass
class InterpolatedRequestOptionsProvider(RequestOptionsProvider):
    """
    Defines the request options to set on an outgoing HTTP request by evaluating `InterpolatedMapping`s

    Attributes:
        config (Config): The user-provided configuration as specified by the source's spec
        request_parameters (Union[str, Mapping[str, str]]): The request parameters to set on an outgoing HTTP request
        request_headers (Union[str, Mapping[str, str]]): The request headers to set on an outgoing HTTP request
        request_body_data (Union[str, Mapping[str, str]]): The body data to set on an outgoing HTTP request
        request_body_json (Union[str, Mapping[str, str]]): The json content to set on an outgoing HTTP request
    """

    parameters: InitVar[Mapping[str, Any]]
    config: Config = field(default_factory=dict)
    request_parameters: Optional[RequestInput] = None
    request_headers: Optional[RequestInput] = None
    request_body: Optional[
        Union[
            RequestBodyGraphQL,
            RequestBodyJsonObject,
            RequestBodyPlainText,
            RequestBodyUrlEncodedForm,
        ]
    ] = None
    request_body_data: Optional[RequestInput] = None
    request_body_json: Optional[NestedMapping] = None
    query_properties_key: Optional[str] = None

    def __post_init__(self, parameters: Mapping[str, Any]) -> None:
        if self.request_parameters is None:
            self.request_parameters = {}
        if self.request_headers is None:
            self.request_headers = {}
        # resolve the request body to either data or json
        self._resolve_request_body()
        # If request_body is not provided, set request_body_data and request_body_json to empty dicts
        if self.request_body_data is None:
            self.request_body_data = {}
        if self.request_body_json is None:
            self.request_body_json = {}
        # If both request_body_data and request_body_json are provided, raise an error
        if self.request_body_json and self.request_body_data:
            raise ValueError(
                "RequestOptionsProvider should only contain either 'request_body_data' or 'request_body_json' not both"
            )
        # set interpolators
        self._parameter_interpolator = InterpolatedRequestInputProvider(
            config=self.config, request_inputs=self.request_parameters, parameters=parameters
        )
        self._headers_interpolator = InterpolatedRequestInputProvider(
            config=self.config, request_inputs=self.request_headers, parameters=parameters
        )
        self._body_data_interpolator = InterpolatedRequestInputProvider(
            config=self.config, request_inputs=self.request_body_data, parameters=parameters
        )
        self._body_json_interpolator = InterpolatedNestedRequestInputProvider(
            config=self.config, request_inputs=self.request_body_json, parameters=parameters
        )

    def _resolve_request_body(self) -> None:
        """
        Resolves the request body configuration by setting either `request_body_data` or `request_body_json`
        based on the type specified in `self.request_body`. If neither is provided, both are initialized as empty
        dictionaries. Raises a ValueError if both `request_body_data` and `request_body_json` are set simultaneously.
        Raises:
            ValueError: if an unsupported request body type is provided.
        """
        # Resolve the request body to either data or json
        if self.request_body is not None and self.request_body.type is not None:
            if self.request_body.type == "RequestBodyUrlEncodedForm":
                self.request_body_data = self.request_body.value
            elif self.request_body.type == "RequestBodyGraphQL":
                self.request_body_json = self.request_body.value.dict(exclude_none=True)
            elif self.request_body.type in ("RequestBodyJsonObject", "RequestBodyPlainText"):
                self.request_body_json = self.request_body.value
            else:
                raise ValueError(f"Unsupported request body type: {self.request_body.type}")

    def get_request_params(
        self,
        *,
        stream_state: Optional[StreamState] = None,
        stream_slice: Optional[StreamSlice] = None,
        next_page_token: Optional[Mapping[str, Any]] = None,
    ) -> MutableMapping[str, Any]:
        interpolated_value = self._parameter_interpolator.eval_request_inputs(
            stream_slice,
            next_page_token,
            valid_key_types=(str,),
            valid_value_types=ValidRequestTypes,
        )
        if isinstance(interpolated_value, dict):
            if self.query_properties_key:
                if not stream_slice:
                    raise ValueError(
                        "stream_slice should not be None if query properties in requests is enabled. Please contact Airbyte Support"
                    )
                elif (
                    "query_properties" not in stream_slice.extra_fields
                    or stream_slice.extra_fields.get("query_properties") is None
                ):
                    raise ValueError(
                        "QueryProperties component is defined but stream_partition does not contain query_properties. Please contact Airbyte Support"
                    )
                elif not isinstance(stream_slice.extra_fields.get("query_properties"), List):
                    raise ValueError(
                        "QueryProperties component is defined but stream_slice.extra_fields.query_properties is not a List. Please contact Airbyte Support"
                    )
                interpolated_value = {
                    **interpolated_value,
                    self.query_properties_key: ",".join(
                        stream_slice.extra_fields.get("query_properties")  # type: ignore  # Earlier type checks validate query_properties type
                    ),
                }
            return interpolated_value
        return {}

    def get_request_headers(
        self,
        *,
        stream_state: Optional[StreamState] = None,
        stream_slice: Optional[StreamSlice] = None,
        next_page_token: Optional[Mapping[str, Any]] = None,
    ) -> Mapping[str, Any]:
        return self._headers_interpolator.eval_request_inputs(stream_slice, next_page_token)

    def get_request_body_data(
        self,
        *,
        stream_state: Optional[StreamState] = None,
        stream_slice: Optional[StreamSlice] = None,
        next_page_token: Optional[Mapping[str, Any]] = None,
    ) -> Union[Mapping[str, Any], str]:
        return self._body_data_interpolator.eval_request_inputs(
            stream_slice,
            next_page_token,
            valid_key_types=(str,),
            valid_value_types=ValidRequestTypes,
        )

    def get_request_body_json(
        self,
        *,
        stream_state: Optional[StreamState] = None,
        stream_slice: Optional[StreamSlice] = None,
        next_page_token: Optional[Mapping[str, Any]] = None,
    ) -> Mapping[str, Any]:
        return self._body_json_interpolator.eval_request_inputs(stream_slice, next_page_token)



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#
import json
import logging
import logging.config
import os
from typing import Any, Callable, Mapping, Optional, Tuple

import orjson

from airbyte_cdk.models import (
    AirbyteLogMessage,
    AirbyteMessage,
    AirbyteMessageSerializer,
    Level,
    Type,
)
from airbyte_cdk.utils import PrintBuffer
from airbyte_cdk.utils.airbyte_secrets_utils import filter_secrets

PRINT_BUFFER = PrintBuffer(flush_interval=0.1)

LOGGING_CONFIG = {
    "version": 1,
    "disable_existing_loggers": False,
    "formatters": {
        "airbyte": {"()": "airbyte_cdk.logger.AirbyteLogFormatter", "format": "%(message)s"},
    },
    "handlers": {
        "console": {
            "class": "logging.StreamHandler",
            "stream": PRINT_BUFFER,
            "formatter": "airbyte",
        },
    },
    "root": {
        "handlers": ["console"],
    },
}


def is_platform_debug_log_enabled() -> bool:
    return os.environ.get("LOG_LEVEL", "info").lower() == "debug"


def init_logger(name: Optional[str] = None) -> logging.Logger:
    """Initial set up of logger"""
    logger = logging.getLogger(name)
    logger.setLevel(logging.INFO)
    logging.config.dictConfig(LOGGING_CONFIG)
    return logger


def lazy_log(logger: logging.Logger, level: int, lazy_log_provider: Callable[[], str]) -> None:
    """
    This method ensure that the processing of the log message is only done if the logger is enabled for the log level.
    """
    if logger.isEnabledFor(level):
        logger.log(level, lazy_log_provider())


class AirbyteLogFormatter(logging.Formatter):
    """Output log records using AirbyteMessage"""

    # Transforming Python log levels to Airbyte protocol log levels
    level_mapping = {
        logging.FATAL: Level.FATAL,
        logging.ERROR: Level.ERROR,
        logging.WARNING: Level.WARN,
        logging.INFO: Level.INFO,
        logging.DEBUG: Level.DEBUG,
    }

    def format(self, record: logging.LogRecord) -> str:
        """Return a JSON representation of the log message"""
        airbyte_level = self.level_mapping.get(record.levelno, "INFO")
        if airbyte_level == Level.DEBUG:
            extras = self.extract_extra_args_from_record(record)
            if is_platform_debug_log_enabled():
                # We have a different behavior between debug logs enabled through `--debug` argument and debug logs
                # enabled through environment variable. The reason is that for platform logs, we need to have these
                # printed as AirbyteMessage which is not the case with the current previous implementation.
                # Why not migrate both to AirbyteMessages then? AirbyteMessages do not support having structured logs.
                # which means that the DX would be degraded compared to the current solution (devs will need to identify
                # the `log.message` field and figure out where in this field is the response while the current solution
                # have a specific field that is structured for extras.
                message = f"{filter_secrets(record.getMessage())} ///\nExtra logs: {filter_secrets(json.dumps(extras))}"
                log_message = AirbyteMessage(
                    type=Type.LOG, log=AirbyteLogMessage(level=airbyte_level, message=message)
                )
                return orjson.dumps(AirbyteMessageSerializer.dump(log_message)).decode()
            else:
                debug_dict = {"type": "DEBUG", "message": record.getMessage(), "data": extras}
                return filter_secrets(json.dumps(debug_dict))
        else:
            message = super().format(record)
            message = filter_secrets(message)
            log_message = AirbyteMessage(
                type=Type.LOG, log=AirbyteLogMessage(level=airbyte_level, message=message)
            )
            return orjson.dumps(AirbyteMessageSerializer.dump(log_message)).decode()

    @staticmethod
    def extract_extra_args_from_record(record: logging.LogRecord) -> Mapping[str, Any]:
        """
        The python logger conflates default args with extra args. We use an empty log record and set operations
        to isolate fields passed to the log record via extra by the developer.
        """
        default_attrs = logging.LogRecord("", 0, "", 0, None, None, None).__dict__.keys()
        extra_keys = set(record.__dict__.keys()) - default_attrs
        return {k: str(getattr(record, k)) for k in extra_keys if hasattr(record, k)}


def log_by_prefix(msg: str, default_level: str) -> Tuple[int, str]:
    """Custom method, which takes log level from first word of message"""
    valid_log_types = ["FATAL", "ERROR", "WARN", "INFO", "DEBUG", "TRACE"]
    split_line = msg.split()
    first_word = next(iter(split_line), None)
    if first_word in valid_log_types:
        log_level = logging.getLevelName(first_word)
        rendered_message = " ".join(split_line[1:])
    else:
        log_level = logging.getLevelName(default_level)
        rendered_message = msg

    return log_level, rendered_message



#
# Copyright (c) 2025 Airbyte, Inc., all rights reserved.
#

from dataclasses import dataclass, field
from typing import Any, List, MutableMapping, Optional, Type, Union

import dpath

from airbyte_cdk.sources.declarative.interpolation.interpolated_boolean import InterpolatedBoolean
from airbyte_cdk.sources.declarative.interpolation.interpolated_string import InterpolatedString
from airbyte_cdk.sources.declarative.transformations.add_fields import (
    AddedFieldDefinition,
    ParsedAddFieldDefinition,
)
from airbyte_cdk.sources.declarative.transformations.config_transformations.config_transformation import (
    ConfigTransformation,
)


@dataclass
class ConfigAddFields(ConfigTransformation):
    """
    Transformation which adds fields to a config. The path of the added field can be nested. Adding nested fields will create all
    necessary parent objects (like mkdir -p).

    This transformation has access to the config being transformed.

    Examples of instantiating this transformation via YAML:
    - type: ConfigAddFields
      fields:
        ### hardcoded constant
        - path: ["path"]
          value: "static_value"

        ### nested path
        - path: ["path", "to", "field"]
          value: "static"

        ### from config
        - path: ["derived_field"]
          value: "{{ config.original_field }}"

        ### by supplying any valid Jinja template directive or expression
        - path: ["two_times_two"]
          value: "{{ 2 * 2 }}"

    Attributes:
        fields (List[AddedFieldDefinition]): A list of transformations (path and corresponding value) that will be added to the config
    """

    fields: List[AddedFieldDefinition]
    condition: str = ""
    _parsed_fields: List[ParsedAddFieldDefinition] = field(
        init=False, repr=False, default_factory=list
    )

    def __post_init__(self) -> None:
        self._filter_interpolator = InterpolatedBoolean(condition=self.condition, parameters={})

        for add_field in self.fields:
            if len(add_field.path) < 1:
                raise ValueError(
                    f"Expected a non-zero-length path for the AddFields transformation {add_field}"
                )

            if not isinstance(add_field.value, InterpolatedString):
                if not isinstance(add_field.value, str):
                    raise ValueError(
                        f"Expected a string value for the AddFields transformation: {add_field}"
                    )
                else:
                    self._parsed_fields.append(
                        ParsedAddFieldDefinition(
                            add_field.path,
                            InterpolatedString.create(add_field.value, parameters={}),
                            value_type=add_field.value_type,
                            parameters={},
                        )
                    )
            else:
                self._parsed_fields.append(
                    ParsedAddFieldDefinition(
                        add_field.path,
                        add_field.value,
                        value_type=add_field.value_type,
                        parameters={},
                    )
                )

    def transform(
        self,
        config: MutableMapping[str, Any],
    ) -> None:
        """
        Transforms a config by adding fields based on the provided field definitions.

        :param config: The user-provided configuration to be transformed
        """
        for parsed_field in self._parsed_fields:
            valid_types = (parsed_field.value_type,) if parsed_field.value_type else None
            value = parsed_field.value.eval(config, valid_types=valid_types)
            if not self.condition or self._filter_interpolator.eval(
                config, value=value, path=parsed_field.path
            ):
                dpath.new(config, parsed_field.path, value)



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#

from typing import Any, Dict


class OneOfOptionConfig:
    """
    Base class to configure a Pydantic model that's used as a oneOf option in a parent model in a way that's compatible with all Airbyte consumers.

    Inherit from this class in the nested Config class in a model and set title and description (these show up in the UI) and discriminator (this is making sure it's marked as required in the schema).

    Usage:

        ```python
        class OptionModel(BaseModel):
            mode: Literal["option_a"] = Field("option_a", const=True)
            option_a_field: str = Field(...)

            class Config(OneOfOptionConfig):
                title = "Option A"
                description = "Option A description"
                discriminator = "mode"
        ```
    """

    @staticmethod
    def schema_extra(schema: Dict[str, Any], model: Any) -> None:
        if hasattr(model.Config, "description"):
            schema["description"] = model.Config.description
        if hasattr(model.Config, "discriminator"):
            schema.setdefault("required", []).append(model.Config.discriminator)



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#

import logging
from abc import ABC
from datetime import timedelta
from typing import Any, Callable, Iterator, List, Mapping, MutableMapping, Optional, Tuple

from airbyte_cdk.models import AirbyteMessage, AirbyteStateMessage, ConfiguredAirbyteCatalog
from airbyte_cdk.sources import AbstractSource
from airbyte_cdk.sources.concurrent_source.concurrent_source import ConcurrentSource
from airbyte_cdk.sources.connector_state_manager import ConnectorStateManager
from airbyte_cdk.sources.streams import Stream
from airbyte_cdk.sources.streams.concurrent.abstract_stream import AbstractStream
from airbyte_cdk.sources.streams.concurrent.abstract_stream_facade import AbstractStreamFacade
from airbyte_cdk.sources.streams.concurrent.adapters import StreamFacade
from airbyte_cdk.sources.streams.concurrent.cursor import (
    ConcurrentCursor,
    Cursor,
    CursorField,
    CursorValueType,
    FinalStateCursor,
    GapType,
)
from airbyte_cdk.sources.streams.concurrent.state_converters.abstract_stream_state_converter import (
    AbstractStreamStateConverter,
)

DEFAULT_LOOKBACK_SECONDS = 0


class ConcurrentSourceAdapter(AbstractSource, ABC):
    def __init__(self, concurrent_source: ConcurrentSource, **kwargs: Any) -> None:
        """
        ConcurrentSourceAdapter is a Source that wraps a concurrent source and exposes it as a regular source.

        The source's streams are still defined through the streams() method.
        Streams wrapped in a StreamFacade will be processed concurrently.
        Other streams will be processed sequentially as a later step.
        """
        self._concurrent_source = concurrent_source
        super().__init__(**kwargs)

    def read(
        self,
        logger: logging.Logger,
        config: Mapping[str, Any],
        catalog: ConfiguredAirbyteCatalog,
        state: Optional[List[AirbyteStateMessage]] = None,
    ) -> Iterator[AirbyteMessage]:
        abstract_streams = self._select_abstract_streams(config, catalog)
        concurrent_stream_names = {stream.name for stream in abstract_streams}
        configured_catalog_for_regular_streams = ConfiguredAirbyteCatalog(
            streams=[
                stream
                for stream in catalog.streams
                if stream.stream.name not in concurrent_stream_names
            ]
        )
        if abstract_streams:
            yield from self._concurrent_source.read(abstract_streams)
        if configured_catalog_for_regular_streams.streams:
            yield from super().read(logger, config, configured_catalog_for_regular_streams, state)

    def _select_abstract_streams(
        self, config: Mapping[str, Any], configured_catalog: ConfiguredAirbyteCatalog
    ) -> List[AbstractStream]:
        """
        Selects streams that can be processed concurrently and returns their abstract representations.
        """
        all_streams = self.streams(config)
        stream_name_to_instance: Mapping[str, Stream] = {s.name: s for s in all_streams}
        abstract_streams: List[AbstractStream] = []
        for configured_stream in configured_catalog.streams:
            stream_instance = stream_name_to_instance.get(configured_stream.stream.name)
            if not stream_instance:
                continue

            if isinstance(stream_instance, AbstractStreamFacade):
                abstract_streams.append(stream_instance.get_underlying_stream())
        return abstract_streams

    def convert_to_concurrent_stream(
        self,
        logger: logging.Logger,
        stream: Stream,
        state_manager: ConnectorStateManager,
        cursor: Optional[Cursor] = None,
    ) -> Stream:
        """
        Prepares a stream for concurrent processing by initializing or assigning a cursor,
        managing the stream's state, and returning an updated Stream instance.
        """
        state: MutableMapping[str, Any] = {}

        if cursor:
            state = state_manager.get_stream_state(stream.name, stream.namespace)

            stream.cursor = cursor  # type: ignore[assignment]  # cursor is of type ConcurrentCursor, which inherits from Cursor
            if hasattr(stream, "parent"):
                stream.parent.cursor = cursor
        else:
            cursor = FinalStateCursor(
                stream_name=stream.name,
                stream_namespace=stream.namespace,
                message_repository=self.message_repository,  # type: ignore[arg-type]  # _default_message_repository will be returned in the worst case
            )
        return StreamFacade.create_from_stream(stream, self, logger, state, cursor)

    def initialize_cursor(
        self,
        stream: Stream,
        state_manager: ConnectorStateManager,
        converter: AbstractStreamStateConverter,
        slice_boundary_fields: Optional[Tuple[str, str]],
        start: Optional[CursorValueType],
        end_provider: Callable[[], CursorValueType],
        lookback_window: Optional[GapType] = None,
        slice_range: Optional[GapType] = None,
    ) -> Optional[ConcurrentCursor]:
        lookback_window = lookback_window or timedelta(seconds=DEFAULT_LOOKBACK_SECONDS)

        cursor_field_name = stream.cursor_field

        if cursor_field_name:
            if not isinstance(cursor_field_name, str):
                raise ValueError(
                    f"Cursor field type must be a string, but received {type(cursor_field_name).__name__}."
                )

            return ConcurrentCursor(
                stream.name,
                stream.namespace,
                state_manager.get_stream_state(stream.name, stream.namespace),
                self.message_repository,  # type: ignore[arg-type]  # _default_message_repository will be returned in the worst case
                state_manager,
                converter,
                CursorField(cursor_field_name),
                slice_boundary_fields,
                start,
                end_provider,
                lookback_window,
                slice_range,
            )

        return None



# Copyright (c) 2023 Airbyte, Inc., all rights reserved.

import contextlib
import functools
from collections import defaultdict
from enum import Enum
from types import TracebackType
from typing import Callable, Dict, Iterable, List, Optional, Union

import requests_mock

from airbyte_cdk.test.mock_http.matcher import HttpRequestMatcher
from airbyte_cdk.test.mock_http.request import HttpRequest
from airbyte_cdk.test.mock_http.response import HttpResponse


class SupportedHttpMethods(str, Enum):
    GET = "get"
    PATCH = "patch"
    POST = "post"
    PUT = "put"
    DELETE = "delete"


class HttpMocker(contextlib.ContextDecorator):
    """
    WARNING 1: This implementation only works if the lib used to perform HTTP requests is `requests`.

    WARNING 2: Given multiple requests that are not mutually exclusive, the request will match the first one. This can happen in scenarios
    where the same request is added twice (in which case there will always be an exception because we will never match the second
    request) or in a case like this:
    ```
    http_mocker.get(HttpRequest(_A_URL, headers={"less_granular": "1", "more_granular": "2"}), <...>)
    http_mocker.get(HttpRequest(_A_URL, headers={"less_granular": "1"}), <...>)
    requests.get(_A_URL, headers={"less_granular": "1", "more_granular": "2"})
    ```
    In the example above, the matcher would match the second mock as requests_mock iterate over the matcher in reverse order (see
    https://github.com/jamielennox/requests-mock/blob/c06f124a33f56e9f03840518e19669ba41b93202/requests_mock/adapter.py#L246) even
    though the request sent is a better match for the first `http_mocker.get`.
    """

    def __init__(self) -> None:
        self._mocker = requests_mock.Mocker()
        self._matchers: Dict[SupportedHttpMethods, List[HttpRequestMatcher]] = defaultdict(list)

    def __enter__(self) -> "HttpMocker":
        self._mocker.__enter__()
        return self

    def __exit__(
        self,
        exc_type: Optional[BaseException],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> None:
        self._mocker.__exit__(exc_type, exc_val, exc_tb)

    def _validate_all_matchers_called(self) -> None:
        for matcher in self._get_matchers():
            if not matcher.has_expected_match_count():
                raise ValueError(f"Invalid number of matches for `{matcher}`")

    def _mock_request_method(
        self,
        method: SupportedHttpMethods,
        request: HttpRequest,
        responses: Union[HttpResponse, List[HttpResponse]],
    ) -> None:
        if isinstance(responses, HttpResponse):
            responses = [responses]

        matcher = HttpRequestMatcher(request, len(responses))
        if matcher in self._matchers[method]:
            raise ValueError(f"Request {matcher.request} already mocked")
        self._matchers[method].append(matcher)

        getattr(self._mocker, method)(
            requests_mock.ANY,
            additional_matcher=self._matches_wrapper(matcher),
            response_list=[
                {
                    self._get_body_field(response): response.body,
                    "status_code": response.status_code,
                    "headers": response.headers,
                }
                for response in responses
            ],
        )

    @staticmethod
    def _get_body_field(response: HttpResponse) -> str:
        return "text" if isinstance(response.body, str) else "content"

    def get(self, request: HttpRequest, responses: Union[HttpResponse, List[HttpResponse]]) -> None:
        self._mock_request_method(SupportedHttpMethods.GET, request, responses)

    def patch(
        self, request: HttpRequest, responses: Union[HttpResponse, List[HttpResponse]]
    ) -> None:
        self._mock_request_method(SupportedHttpMethods.PATCH, request, responses)

    def post(
        self, request: HttpRequest, responses: Union[HttpResponse, List[HttpResponse]]
    ) -> None:
        self._mock_request_method(SupportedHttpMethods.POST, request, responses)

    def put(self, request: HttpRequest, responses: Union[HttpResponse, List[HttpResponse]]) -> None:
        self._mock_request_method(SupportedHttpMethods.PUT, request, responses)

    def delete(
        self, request: HttpRequest, responses: Union[HttpResponse, List[HttpResponse]]
    ) -> None:
        self._mock_request_method(SupportedHttpMethods.DELETE, request, responses)

    @staticmethod
    def _matches_wrapper(
        matcher: HttpRequestMatcher,
    ) -> Callable[[requests_mock.request._RequestObjectProxy], bool]:
        def matches(requests_mock_request: requests_mock.request._RequestObjectProxy) -> bool:
            # query_params are provided as part of `requests_mock_request.url`
            http_request = HttpRequest(
                requests_mock_request.url,
                query_params={},
                headers=requests_mock_request.headers,
                body=requests_mock_request.body,
            )
            return matcher.matches(http_request)

        return matches

    def assert_number_of_calls(self, request: HttpRequest, number_of_calls: int) -> None:
        corresponding_matchers = list(
            filter(lambda matcher: matcher.request is request, self._get_matchers())
        )
        if len(corresponding_matchers) != 1:
            raise ValueError(
                f"Was expecting only one matcher to match the request but got `{corresponding_matchers}`"
            )

        assert corresponding_matchers[0].actual_number_of_matches == number_of_calls

    # trying to type that using callables provides the error `incompatible with return type "_F" in supertype "ContextDecorator"`
    def __call__(self, f):  # type: ignore
        @functools.wraps(f)
        def wrapper(*args, **kwargs):  # type: ignore  # this is a very generic wrapper that does not need to be typed
            with self:
                assertion_error = None

                kwargs["http_mocker"] = self
                try:
                    result = f(*args, **kwargs)
                except requests_mock.NoMockAddress as no_mock_exception:
                    matchers_as_string = "\n\t".join(
                        map(lambda matcher: str(matcher.request), self._get_matchers())
                    )
                    raise ValueError(
                        f"No matcher matches {no_mock_exception.args[0]} with headers `{no_mock_exception.request.headers}` "
                        f"and body `{no_mock_exception.request.body}`. "
                        f"Matchers currently configured are:\n\t{matchers_as_string}."
                    ) from no_mock_exception
                except AssertionError as test_assertion:
                    assertion_error = test_assertion

                # We validate the matchers before raising the assertion error because we want to show the tester if an HTTP request wasn't
                # mocked correctly
                try:
                    self._validate_all_matchers_called()
                except ValueError as http_mocker_exception:
                    # This seems useless as it catches ValueError and raises ValueError but without this, the prevailing error message in
                    # the output is the function call that failed the assertion, whereas raising `ValueError(http_mocker_exception)`
                    # like we do here provides additional context for the exception.
                    raise ValueError(http_mocker_exception) from None
                if assertion_error:
                    raise assertion_error
                return result

        return wrapper

    def _get_matchers(self) -> Iterable[HttpRequestMatcher]:
        for matchers in self._matchers.values():
            yield from matchers

    def clear_all_matchers(self) -> None:
        """Clears all stored matchers by resetting the _matchers list to an empty state."""
        self._matchers = defaultdict(list)



#
# Copyright (c) 2025 Airbyte, Inc., all rights reserved.
#

from airbyte_cdk.sources.declarative.validators.dpath_validator import DpathValidator
from airbyte_cdk.sources.declarative.validators.predicate_validator import PredicateValidator
from airbyte_cdk.sources.declarative.validators.validate_adheres_to_schema import (
    ValidateAdheresToSchema,
)
from airbyte_cdk.sources.declarative.validators.validation_strategy import ValidationStrategy
from airbyte_cdk.sources.declarative.validators.validator import Validator

__all__ = [
    "Validator",
    "DpathValidator",
    "ValidationStrategy",
    "ValidateAdheresToSchema",
    "PredicateValidator",
]
